import abc
import logging
import ssl
import typing
import warnings
from datetime import timedelta, timezone, tzinfo
from pathlib import Path as _Path
from typing import (
    Any,
    Callable,
    ClassVar,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from mode import Seconds as _Seconds, want_seconds
from mode.utils.imports import SymbolArg, symbol_by_name
from mode.utils.logging import Severity as _Severity
from yarl import URL as _URL

from faust.exceptions import ImproperlyConfigured
from faust.types.auth import CredentialsArg, CredentialsT, to_credentials
from faust.types.codecs import CodecArg, CodecT
from faust.utils import json
from faust.utils.urls import URIListArg, urllist

if typing.TYPE_CHECKING:
    from .sections import Section as _Section
    from .settings import Settings as _Settings
else:

    class _Section:
        ...  # noqa

    class _Settings:
        ...  # noqa


__all__ = [
    "AutodiscoverArg",
    "DictArg",
    "URLArg",
    "BrokerArg",
    "Param",
    "Bool",
    "Str",
    "Severity",
    "Int",
    "UnsignedInt",
    "Version",
    "Port",
    "Seconds",
    "Credentials",
    "SSLContext",
    "Dict",
    "LogHandlers",
    "Timezone",
    "BrokerList",
    "URL",
    "Path",
    "Codec",
    "Enum",
    "Symbol",
    "to_bool",
]

#: Default transport used when no scheme specified.
DEFAULT_BROKER_SCHEME = "kafka"

T = TypeVar("T")
IT = TypeVar("IT")  # Input type.
OT = TypeVar("OT")  # Output type.

BOOLEAN_TERMS: Mapping[str, bool] = {
    "": False,
    "false": False,
    "no": False,
    "0": False,
    "true": True,
    "yes": True,
    "1": True,
    "on": True,
    "off": False,
}

AutodiscoverArg = Union[
    bool,
    Iterable[str],
    Callable[[], Iterable[str]],
]

DictArg = Union[str, Mapping[str, T]]

URLArg = Union[str, _URL]
BrokerArg = URIListArg

DEPRECATION_WARNING_TEMPLATE = """
Setting {self.name} is deprecated since Faust version \
{self.version_deprecated}: {self.deprecation_reason}. {alt_removal}
""".strip()

DEPRECATION_REMOVAL_WARNING = """
Further the setting is scheduled to be removed in Faust version \
{self.version_removal}.
""".strip()


def to_bool(
    term: Union[str, bool], *, table: Mapping[str, bool] = BOOLEAN_TERMS
) -> bool:
    """Convert common terms for true/false to bool.

    Examples (true/false/yes/no/on/off/1/0).
    """
    if table is None:
        table = BOOLEAN_TERMS
    if isinstance(term, str):
        try:
            return table[term.lower()]
        except KeyError:
            raise TypeError("Cannot coerce {0!r} to type bool".format(term))
    return term


OutputCallable = Callable[[_Settings, OT], OT]
OnDefaultCallable = Callable[[_Settings], IT]


class Param(Generic[IT, OT], property):
    """Faust setting desscription.

    Describes a Faust setting, how to read it from environment
    variables or from a configuration object.

    """

    #: Textual description of setting type.
    #: This is used by :file:`extra/tools/render_configuration_reference.py`
    #: to display the types supported by this setting.
    #:
    #: Can be a tuple of actual classes, or a tuple of strings.
    #: If a tuple of classes say a setting that accepts :class:`str`
    #: and :class:`int`:
    #:
    #: .. sourcecode:: python
    #:
    #:      text_type = (str, int)
    #:
    #: the generated description will be:
    #:
    #: .. sourcecode:: restructuredtext
    #:
    #:   :type: :class:`str` / :class:`int`
    #:
    text_type: ClassVar[Tuple[Any, ...]] = (Any,)

    #: Setting name (e.g. ``broker_request_timeout``).
    name: str

    #: Storage name (e.g. ``_broker_request_timeout``).
    #: This is the attribute name where we'll be storing
    #: the actual value for the setting.
    #: For example ``Settings.broker_request_timeout`` will be a
    #: property that calls ``Param.__get__`` on attribute access,
    #: and ``.__set__`` when setting attribute value, and those
    #: will use the underlying ``Settings._broker_request_timeout``
    #: storage attribute to access/store the current value.
    storage_name: str

    #: Default value for setting.
    #:
    #: Note that this will not be used if :attr:`default_alias` or
    #: :attr:`defaut_template` is set.
    default: IT = cast(IT, None)

    #: Environment variable name for setting.
    #: For :setting:`broker_request_timeout` this would be
    #: ``env_name="BROKER_REQUEST_TIMEOUT"``
    env_name: Optional[str] = None

    #: If setting is not customized this is an optional
    #: list of other settings that we should take default value from.
    #: For example the :setting:`broker_consumer`/:setting:`broker_producer`
    #: settings can configure the broker URL for consumers and producers
    #: separately but take their default value from the `broker` setting.
    default_alias: Optional[str] = None

    #: Default template.
    #: If set the default value will be generated from this format string
    #: template.
    #: For exmaple the :setting:`canonical_url` setting uses
    #: ``default_template='http://{conf.web_host}:{conf.web_port}' to
    #: generate a default value from the :setting:`web_host` and
    #: :setting:`web_port` settings.
    default_template: Optional[str] = None

    #: Set to true if the value can be :const:`None`.
    allow_none: bool = False

    # If set to True we don't modify the value
    # of the attribute to set a default.
    # This is used by e.g. the env_prefix setting
    # which has custom constructor code in Settings.on_init
    ignore_default: bool = False

    #: The configuration section this setting belongs to.
    #: E.g. ``sections.Common``.
    section: _Section

    #: The version that this setting was first introduced.
    #: This is used by :file:`extra/tools/render_configuration_reference.py`
    #: to generate a version added directive:
    #:
    #: .. sourcecode:: restructuredtext
    #:
    #:    .. versionadded:: 1.10
    version_introduced: Optional[str] = None

    #: Set this if the setting is deprecated and should not be used anymore.
    #: Deprecated settings are not added to the configuration reference.
    #: Note: You must also set a :attr:`deprecation_reason`.
    version_deprecated: Optional[str] = None
    deprecation_reason: Optional[str] = None

    #: Mapping of version changes and reason for changing.
    #: This is used by :file:`extra/tools/render_configuration_reference.py`
    #: For example if this was enabled by default but then changed
    #: to be disabled by default in version 1.30, then you can specify
    #: that as ``version_changed={'1.30': 'Disabled by default.'}``
    #: and the configuration reference will be rendered with the
    #: following version changed directive added:
    #:
    #: .. sourcecode:: restructuredtext
    #:
    #:    .. versionchanged:: 1.30
    #:
    #:         Disabled by default.
    version_changed: Optional[Mapping[str, str]] = None

    #: Set this if the setting should be disabled completely,
    #: but still be included in the code.
    #: This is rare, no setting should be included in the
    #: code if it has been removed.  Currently this is only
    #: used for the example setting that describes how you can add new
    #: settings.
    version_removed: Optional[str] = None

    #: Mapping of related command line options.
    #: This should be a mapping from command name to a list of option names.
    #:
    #: For example the :setting:`canonical_url` setting lists related
    #: options as:
    #:
    #: .. sourcecode:: python
    #:
    #:    related_cli_options={
    #:      'faust worker': ['--web-host', '--web-port'],
    #:    }
    #:
    #: And this will end up in the configuration reference as:
    #:
    #: .. sourcecode:: restructuredtext
    #:
    #:   :related-options: :option:`faust worker --web-host`,
    #:                     :option:`faust worker --web-port`
    related_cli_options: Mapping[str, List[str]]

    #: List of related settings.
    #: For example for the :setting:`canonical_url` setting
    #: the list of related settings are defined as:
    #: ``related_settings=[web_host, web_port]``.
    #: The configuration reference will then include that as:
    #:
    #: .. sourcecode:: restructuredtext
    #:
    #:    :related-settings: :setting:`web_host`, :setting:`web_port`
    related_settings: List[Any]

    #: Template used to generate a deprecation warning for deprecated settings.
    deprecation_warning_template: str = DEPRECATION_WARNING_TEMPLATE

    #: Template used to generate an additional removal warning
    #: for the deprecation warning.
    deprecation_removal_warning: str = DEPRECATION_REMOVAL_WARNING

    def __init__(
        self,
        *,
        name: str,
        env_name: str = None,
        default: IT = None,
        default_alias: str = None,
        default_template: str = None,
        allow_none: bool = None,
        ignore_default: bool = None,
        section: _Section = None,
        version_introduced: str = None,
        version_deprecated: str = None,
        version_removed: str = None,
        version_changed: Mapping[str, str] = None,
        deprecation_reason: str = None,
        related_cli_options: Mapping[str, List[str]] = None,
        related_settings: List[Any] = None,
        help: str = None,
        **kwargs: Any,
    ) -> None:
        assert name
        self.name = name
        self.storage_name = f"_{name}"
        if env_name is not None:
            self.env_name = env_name
        if default is not None:
            self.default = default
        if default_alias is not None:
            self.default_alias = default_alias
        if default_template is not None:
            self.default_template = default_template
        if allow_none is not None:
            self.allow_none = allow_none
        if ignore_default is not None:
            self.ignore_default = ignore_default
        if section is not None:
            self.section = section
        assert self.section
        if version_introduced is not None:
            self.version_introduced = version_introduced
        if version_deprecated is not None:
            self.version_deprecated = version_deprecated
        if version_removed is not None:
            self.version_removed = version_removed
        if version_changed is not None:
            self.version_changed = version_changed
        if deprecation_reason is not None:
            self.deprecation_reason = deprecation_reason
        if help is not None:
            self.__doc__ = help
        self._on_get_value_: Optional[OutputCallable] = None
        self._on_set_default_: Optional[OnDefaultCallable] = None
        self.options = kwargs
        self.related_cli_options = related_cli_options or {}
        self.related_settings = related_settings or []
        self._init_options(**self.options)

        if self.version_deprecated:
            assert self.deprecation_reason

    def _init_options(self, **kwargs: Any) -> None:
        """Use in subclasses to quickly override ``__init__``."""
        ...

    def on_get_value(self, fun: OutputCallable) -> OutputCallable:
        """Decorator that adds a callback when this setting is retrieved."""
        assert self._on_get_value_ is None
        self._on_get_value_ = fun
        return fun

    def on_set_default(self, fun: OnDefaultCallable) -> OnDefaultCallable:
        """Decorator that adds a callback when a default value is used."""
        assert self._on_set_default_ is None
        self._on_set_default_ = fun
        return fun

    def __get__(self, obj: Any, type: Type = None) -> OT:
        if obj is None:
            return self  # type: ignore
        if self.version_deprecated:
            # we use UserWarning because DeprecationWarning is silenced
            # by default in Python.
            warnings.warn(UserWarning(self.build_deprecation_warning()))
        return self.on_get(obj)

    def __set__(self, obj: Any, value: IT) -> None:
        self.on_set(obj, self.prepare_set(obj, value))

    def on_get(self, conf: _Settings) -> OT:
        """What happens when the setting is accessed/retrieved."""
        value = getattr(conf, self.storage_name)
        if value is None and self.default_alias:
            retval = getattr(conf, self.default_alias)
        else:
            retval = self.prepare_get(conf, value)
        if self._on_get_value_ is not None:
            return self._on_get_value_(conf, retval)
        return retval

    def prepare_get(self, conf: _Settings, value: OT) -> OT:
        """Prepare value when accessed/retrieved."""
        return value

    def on_set(self, settings: Any, value: OT) -> None:
        """What happens when the setting is stored/set."""
        settings.__dict__[self.storage_name] = value
        assert getattr(settings, self.storage_name) == value

    def set_class_default(self, cls: Type) -> None:
        """Set class default value for storage attribute."""
        setattr(cls, self.storage_name, self.default)

    def on_init_set_value(self, conf: _Settings, provided_value: Optional[IT]) -> None:
        """What happens at ``Settings.__init__`` to store provided value.

        Arguments:
            conf: Settings object.
            provided_value: Provided configuration value passed to
                            ``Settings.__init__`` or :const:`None` if not set.
        """
        if provided_value is not None:
            self.__set__(conf, provided_value)

    def on_init_set_default(
        self, conf: _Settings, provided_value: Optional[IT]
    ) -> None:
        """What happens at ``Settings.__init__`` to set default value.

        Arguments:
            conf: Settings object.
            provided_value: Provided configuration value passed to
                            ``Settings.__init__`` or :const:`None` if not set.
        """
        if provided_value is None:
            default_value = self.default
            if self._on_set_default_:
                default_value = self._on_set_default_(conf)
            if default_value is None and self.default_template:
                default_value = self.default_template.format(conf=conf)
            setattr(
                conf, self.storage_name, self.prepare_init_default(conf, default_value)
            )

    def build_deprecation_warning(self) -> str:
        """Build deprecation warning for this setting."""
        alt_removal = ""
        if self.version_removed:
            alt_removal = self.deprecation_removal_warning.format(self=self)
        return self.deprecation_warning_template.format(
            self=self,
            alt_removal=alt_removal,
        )

    def validate_before(self, value: IT = None) -> None:
        """Validate value before setting is converted to the target type."""
        ...

    def validate_after(self, value: OT) -> None:
        """Validate value after it has been converted to its target type."""
        ...

    def prepare_set(self, conf: _Settings, value: IT) -> OT:
        """Prepare value for storage."""
        skip_validate = value is None and self.allow_none
        if not skip_validate:
            self.validate_before(value)
        if value is not None:
            new_value = self.to_python(conf, value)
        else:
            new_value = value
        if not skip_validate:
            self.validate_after(new_value)
        return new_value

    def prepare_init_default(self, conf: _Settings, value: IT) -> OT:
        """Prepare default value for storage."""
        if value is not None:
            return self.to_python(conf, value)
        return None

    def to_python(self, conf: _Settings, value: IT) -> OT:
        """Convert value in input type to its output type."""
        return cast(OT, value)

    @property
    def active(self) -> bool:
        return not bool(self.version_removed)

    @property
    def deprecated(self) -> bool:
        return bool(self.version_deprecated)

    @property
    def class_name(self) -> str:
        return type(self).__name__


class Bool(Param[Any, bool]):
    """Boolean setting type."""

    text_type = (bool,)

    def to_python(self, conf: _Settings, value: Any) -> bool:
        """Convert given value to :class:`bool`."""
        if isinstance(value, str):
            return to_bool(value)
        return bool(value)


class Str(Param[str, str]):
    """String setting type."""

    text_type = (str,)


class Severity(Param[_Severity, _Severity]):
    """Logging severity setting type."""

    text_type = (str, int)


class Number(Param[IT, OT]):
    """Number setting type (baseclass for int/float)."""

    min_value: Optional[int] = None
    max_value: Optional[int] = None
    number_aliases: Mapping[IT, OT]

    def _init_options(
        self,
        min_value: int = None,
        max_value: int = None,
        number_aliases: Mapping[IT, OT] = None,
        **kwargs: Any,
    ) -> None:
        if min_value is not None:
            self.min_value = min_value
        if max_value is not None:
            self.max_value = max_value
        self.number_aliases = number_aliases or {}

    @abc.abstractmethod
    def convert(self, conf: _Settings, value: IT) -> OT:
        ...

    def to_python(self, conf: _Settings, value: IT) -> OT:
        """Convert given value to number."""
        try:
            return self.number_aliases[value]
        except KeyError:
            return self.convert(conf, value)

    def validate_after(self, value: OT) -> None:
        """Validate number value."""
        v = cast(int, value)
        min_ = self.min_value
        max_ = self.max_value
        if min_ is not None and v < min_:
            raise self._out_of_range(v)
        if max_ is not None and v > max_:
            raise self._out_of_range(v)

    def _out_of_range(self, value: float) -> ImproperlyConfigured:
        return ImproperlyConfigured(
            f"Value {value} is out of range for {self.class_name} "
            f"(min={self.min_value} max={self.max_value})"
        )


NumberInputArg = Union[str, int, float]


class _Int(Number[IT, OT]):
    text_type = (int,)

    def convert(self, conf: _Settings, value: IT) -> OT:
        """Convert given value to int."""
        return cast(OT, int(cast(int, value)))


class Int(_Int[NumberInputArg, int]):
    """Signed integer setting type."""


class UnsignedInt(_Int[NumberInputArg, int]):
    """Unsigned integer setting type."""

    min_value = 0


class Version(Int):
    """Version setting type.

    Versions must be greater than ``1``.
    """

    min_value = 1


class Port(UnsignedInt):
    """Network port setting type.

    Ports must be in the range 1-65535.
    """

    min_value = 1
    max_value = 65535


class Seconds(Param[_Seconds, float]):
    """Seconds setting type.

    Converts from :class:`float`/:class:`~datetime.timedelta` to
    :class:`float`.
    """

    text_type = (float, timedelta)

    def to_python(self, conf: _Settings, value: _Seconds) -> float:
        return want_seconds(value)


class Credentials(Param[CredentialsArg, Optional[CredentialsT]]):
    """Authentication credentials setting type."""

    text_type = (CredentialsT,)

    def to_python(
        self, conf: _Settings, value: CredentialsArg
    ) -> Optional[CredentialsT]:
        return to_credentials(value)


class SSLContext(Param[ssl.SSLContext, Optional[ssl.SSLContext]]):
    """SSL context setting type."""

    text_type = (ssl.SSLContext,)


class Dict(Param[DictArg[T], Mapping[str, T]]):
    """Dictionary setting type."""

    text_type = (dict,)

    def to_python(self, conf: _Settings, value: DictArg[T]) -> Mapping[str, T]:
        if isinstance(value, str):
            return json.loads(value)
        elif isinstance(value, Mapping):
            return value
        return dict(value)


class LogHandlers(Param[List[logging.Handler], List[logging.Handler]]):
    """Log handler list setting type."""

    text_type = (List[logging.Handler],)

    def prepare_init_default(
        self, conf: _Settings, value: Any
    ) -> List[logging.Handler]:
        return []


class Timezone(Param[Union[str, tzinfo], tzinfo]):
    """Timezone setting type."""

    text_type = (tzinfo,)
    builtin_timezones = {"UTC": timezone.utc}

    def to_python(self, conf: _Settings, value: Union[str, tzinfo]) -> tzinfo:
        if isinstance(value, str):
            try:
                return cast(tzinfo, self.builtin_timezones[value.lower()])
            except KeyError:
                import pytz

                return cast(tzinfo, pytz.timezone(value))
        else:
            return value


class BrokerList(Param[BrokerArg, List[_URL]]):
    """Broker URL list setting type."""

    text_type = (str, _URL, List[str])
    default_scheme = DEFAULT_BROKER_SCHEME

    def to_python(self, conf: _Settings, value: BrokerArg) -> List[_URL]:
        return self.broker_list(value)

    def broker_list(self, value: BrokerArg) -> List[_URL]:
        return urllist(value, default_scheme=self.default_scheme)


class URL(Param[URLArg, _URL]):
    """URL setting type."""

    text_type = (str, _URL)

    def to_python(self, conf: _Settings, value: URLArg) -> _URL:
        return _URL(value)


class Path(Param[Union[str, _Path], _Path]):
    """Path setting type."""

    text_type = (str, _Path)
    expanduser: bool = True

    def to_python(self, conf: _Settings, value: Union[str, _Path]) -> _Path:
        p = _Path(value)
        if self.expanduser:
            p = p.expanduser()
        return self.prepare_path(conf, p)

    def prepare_path(self, conf: _Settings, path: _Path) -> _Path:
        return path


class Codec(Param[CodecArg, CodecArg]):
    """Serialization codec setting type."""

    text_type = (str, CodecT)


def Enum(typ: T) -> Type[Param[Union[str, T], T]]:
    """Generate new enum setting type."""

    class EnumParam(Param[Union[str, T], T]):
        text_type = (str,)

        def to_python(self, conf: _Settings, value: Union[str, T]) -> T:
            return typ(value)  # type: ignore

    return EnumParam


class _Symbol(Param[IT, OT]):
    text_type = (str, Type)

    def to_python(self, conf: _Settings, value: IT) -> OT:
        return cast(OT, symbol_by_name(value))


def Symbol(typ: T) -> Type[Param[SymbolArg[T], T]]:
    """Generate new symbol setting type."""
    return _Symbol[SymbolArg[T], T]



import abc
import asyncio
import typing
from typing import Any, Mapping, Optional, Pattern, Sequence, Set, Union

from mode import Seconds
from mode.utils.queues import ThrowableQueue

from .channels import ChannelT
from .codecs import CodecArg
from .tuples import TP

if typing.TYPE_CHECKING:
    from .app import AppT as _AppT
    from .models import ModelArg as _ModelArg
    from .serializers import SchemaT as _SchemaT
else:

    class _AppT:
        ...  # noqa

    class _ModelArg:
        ...  # noqa

    class _SchemaT:
        ...  # noqa


__all__ = ["TopicT"]


class TopicT(ChannelT):

    #: Iterable/Sequence of topic names to subscribe to.
    topics: Sequence[str]

    #: Topic retention setting: expiry time in seconds
    #: for messages in the topic.
    retention: Optional[Seconds]

    #: Flag that when enabled means the topic can be "compacted":
    #: if the topic is a log of key/value pairs, the broker can delete
    #: old values for the same key.
    compacting: Optional[bool]

    deleting: Optional[bool]

    #: Number of replicas for topic.
    replicas: Optional[int]

    #: Additional configuration as a mapping.
    config: Optional[Mapping[str, Any]]

    #: Enable acks for this topic.
    acks: bool

    #: Mark topic as internal: it's owned by us and we are allowed
    #: to create or delete the topic as necessary.
    internal: bool

    has_prefix: bool = False

    active_partitions: Optional[Set[TP]]

    @abc.abstractmethod
    def __init__(
        self,
        app: _AppT,
        *,
        topics: Sequence[str] = None,
        pattern: Union[str, Pattern] = None,
        schema: _SchemaT = None,
        key_type: _ModelArg = None,
        value_type: _ModelArg = None,
        is_iterator: bool = False,
        partitions: int = None,
        retention: Seconds = None,
        compacting: bool = None,
        deleting: bool = None,
        replicas: int = None,
        acks: bool = True,
        internal: bool = False,
        config: Mapping[str, Any] = None,
        queue: ThrowableQueue = None,
        key_serializer: CodecArg = None,
        value_serializer: CodecArg = None,
        maxsize: int = None,
        root: ChannelT = None,
        active_partitions: Set[TP] = None,
        allow_empty: bool = False,
        has_prefix: bool = False,
        loop: asyncio.AbstractEventLoop = None
    ) -> None:
        ...

    @property
    @abc.abstractmethod
    def pattern(self) -> Optional[Pattern]:
        ...

    @pattern.setter
    def pattern(self, pattern: Union[str, Pattern]) -> None:
        ...

    @property
    @abc.abstractmethod
    def partitions(self) -> Optional[int]:
        ...

    @partitions.setter
    def partitions(self, partitions: int) -> None:
        ...

    @abc.abstractmethod
    def derive(self, **kwargs: Any) -> ChannelT:
        ...

    @abc.abstractmethod
    def derive_topic(
        self,
        *,
        topics: Sequence[str] = None,
        schema: _SchemaT = None,
        key_type: _ModelArg = None,
        value_type: _ModelArg = None,
        partitions: int = None,
        retention: Seconds = None,
        compacting: bool = None,
        deleting: bool = None,
        internal: bool = False,
        config: Mapping[str, Any] = None,
        prefix: str = "",
        suffix: str = "",
        **kwargs: Any
    ) -> "TopicT":
        ...



"""Agent replies: waiting for replies, sending them, etc."""
import asyncio
from collections import defaultdict
from typing import Any, AsyncIterator, MutableMapping, MutableSet, NamedTuple, Optional
from weakref import WeakSet

from mode import Service

from faust.types import AppT, ChannelT, TopicT

from .models import ReqRepResponse

__all__ = ["ReplyPromise", "BarrierState", "ReplyConsumer"]


class ReplyTuple(NamedTuple):
    correlation_id: str
    value: Any


class ReplyPromise(asyncio.Future):
    """Reply promise can be :keyword:`await`-ed to wait until result ready."""

    reply_to: str
    correlation_id: str

    def __init__(self, reply_to: str, correlation_id: str = "", **kwargs: Any) -> None:
        self.reply_to = reply_to
        self._verify_correlation_id(correlation_id)
        self.correlation_id = correlation_id
        self.__post_init__()
        super().__init__(**kwargs)

    def __post_init__(self) -> None:
        ...

    def _verify_correlation_id(self, correlation_id: str) -> None:
        if not correlation_id:
            raise ValueError("ReplyPromise missing correlation_id argument.")

    def fulfill(self, correlation_id: str, value: Any) -> None:
        """Fulfill promise: a reply was received."""
        # If it wasn't for BarrierState we would just use .set_result()
        # directly, but BarrierState.fulfill requires the correlation_id
        # to be sent with it. That way it can mark that part of the map
        # operation as completed.
        assert correlation_id == self.correlation_id
        self.set_result(value)


class BarrierState(ReplyPromise):
    """State of pending/complete barrier.

    A barrier is a synchronization primitive that will wait until
    a group of coroutines have completed.
    """

    #: This is the size while the messages are being sent.
    #: (it's a tentative total, added to until the total is finalized).
    size: int = 0

    #: This is the actual total when all messages have been sent.
    #: It's set by :meth:`finalize`.
    total: int = 0

    #: The number of results we have received.
    fulfilled: int = 0

    #: Internal queue where results are added to.
    _results: asyncio.Queue

    #: Set of pending replies that this barrier is composed of.
    pending: MutableSet[ReplyPromise]

    def __post_init__(self) -> None:
        self.pending = set()
        loop: asyncio.AbstractEventLoop = self._loop
        self._results = asyncio.Queue(maxsize=1000, loop=loop)

    def _verify_correlation_id(self, correlation_id: str) -> None:
        pass  # barrier does not require a correlation id.

    def add(self, p: ReplyPromise) -> None:
        """Add promise to barrier.

        Note:
            You can only add promises before the barrier is finalized
            using :meth:`finalize`.
        """
        self.pending.add(p)
        self.size += 1

    def finalize(self) -> None:
        """Finalize this barrier.

        After finalization you can not grow or shrink the size
        of the barrier.
        """
        self.total = self.size
        # The barrier may have been filled up already at this point,
        if self.fulfilled >= self.total:
            self.set_result(True)
            self._results.put_nowait(None)  # always wake-up .iterate()

    def fulfill(self, correlation_id: str, value: Any) -> None:
        """Fulfill one of the promises in this barrier.

        Once all promises in this barrier is fulfilled, the barrier
        will be ready.
        """
        # ReplyConsumer calls this whenever a new reply is received.
        self._results.put_nowait(ReplyTuple(correlation_id, value))
        self.fulfilled += 1
        if self.total:
            if self.fulfilled >= self.total:
                self.set_result(True)
                self._results.put_nowait(None)  # always wake-up .iterate()

    def get_nowait(self) -> ReplyTuple:
        """Return next reply, or raise :exc:`asyncio.QueueEmpty`."""
        for _ in range(10):  # remove sentinels
            value = self._results.get_nowait()
            if value is not None:
                return value
        raise asyncio.QueueEmpty()

    async def iterate(self) -> AsyncIterator[ReplyTuple]:
        """Iterate over results as they arrive."""
        get = self._results.get
        get_nowait = self._results.get_nowait
        is_done = self.done
        while not is_done():
            value = await get()
            if value is not None:
                yield value
        while 1:
            try:
                value = get_nowait()
            except asyncio.QueueEmpty:
                break
            else:
                if value is not None:
                    yield value


class ReplyConsumer(Service):
    """Consumer responsible for redelegation of replies received."""

    _waiting: MutableMapping[str, MutableSet[ReplyPromise]]
    _fetchers: MutableMapping[str, Optional[asyncio.Future]]

    def __init__(self, app: AppT, **kwargs: Any) -> None:
        self.app = app
        self._waiting = defaultdict(WeakSet)
        self._fetchers = {}
        super().__init__(**kwargs)

    async def on_start(self) -> None:
        """Call when reply consumer starts."""
        if self.app.conf.reply_create_topic:
            await self._start_fetcher(self.app.conf.reply_to)

    async def add(self, correlation_id: str, promise: ReplyPromise) -> None:
        """Register promise to start tracking when it arrives."""
        reply_topic = promise.reply_to
        if reply_topic not in self._fetchers:
            await self._start_fetcher(reply_topic)
        self._waiting[correlation_id].add(promise)

    async def _start_fetcher(self, topic_name: str) -> None:
        if topic_name not in self._fetchers:
            # set the key as a lock, so it doesn't happen twice
            self._fetchers[topic_name] = None
            # declare the topic
            topic = self._reply_topic(topic_name)
            await topic.maybe_declare()
            await self.sleep(3.0)
            # then create the future
            self._fetchers[topic_name] = self.add_future(self._drain_replies(topic))

    async def _drain_replies(self, channel: ChannelT) -> None:
        async for reply in channel.stream():
            for promise in self._waiting[reply.correlation_id]:
                promise.fulfill(reply.correlation_id, reply.value)

    def _reply_topic(self, topic: str) -> TopicT:
        return self.app.topic(
            topic,
            partitions=1,
            replicas=0,
            deleting=True,
            retention=self.app.conf.reply_expires,
            value_type=ReqRepResponse,
        )



"""The conductor delegates messages from the consumer to the streams."""
import asyncio
import os
import typing
from collections import defaultdict
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    MutableMapping,
    MutableSet,
    Optional,
    Set,
    Tuple,
    cast,
)

from mode import Service, get_logger
from mode.utils.futures import notify

from faust.exceptions import KeyDecodeError, ValueDecodeError
from faust.types import TP, AppT, EventT, K, Message, V
from faust.types.topics import TopicT
from faust.types.transports import ConductorT, ConsumerCallback, TPorTopicSet
from faust.types.tuples import tp_set_to_map
from faust.utils.tracing import traced_from_parent_span

if typing.TYPE_CHECKING:  # pragma: no cover
    from faust.topics import Topic as _Topic
else:

    class _Topic:
        ...  # noqa


NO_CYTHON = bool(os.environ.get("NO_CYTHON", False))

if not NO_CYTHON:  # pragma: no cover
    try:
        from ._cython.conductor import ConductorHandler
    except ImportError:
        ConductorHandler = None
else:  # pragma: no cover
    ConductorHandler = None

__all__ = ["Conductor", "ConductorCompiler"]

logger = get_logger(__name__)


class ConductorCompiler:  # pragma: no cover
    """Compile a function to handle the messages for a topic+partition."""

    def build(
        self, conductor: "Conductor", tp: TP, channels: MutableSet[_Topic]
    ) -> ConsumerCallback:
        """Generate closure used to deliver messages."""
        # This method localizes variables and attribute access
        # for better performance.  This is part of the inner loop
        # of a Faust worker, so tiny improvements here has big impact.

        topic, partition = tp
        app = conductor.app
        len_: Callable[[Any], int] = len

        # We divide `stream_buffer_maxsize` with Queue.pressure_ratio
        # find a limit to the number of messages we will buffer
        # before considering the buffer to be under high pressure.
        # When the buffer is under high pressure, we call
        # Consumer.on_buffer_full(tp) to remove this topic partition
        # from the fetcher.
        # We still accept anything that's currently in the fetcher (it's
        # already in memory so we are just moving the data) without blocking,
        # but signal the fetcher to stop retrieving any more data for this
        # partition.
        consumer_on_buffer_full = app.consumer.on_buffer_full

        # when the buffer drops down to half we re-enable fetching
        # from the partition.
        consumer_on_buffer_drop = app.consumer.on_buffer_drop

        # flow control will completely block the streams from processing
        # more data, and is used during rebalancing.
        acquire_flow_control: Callable = app.flow_control.acquire

        # when streams send new messages as a side effect, the producer
        # buffer can sometimes fill up, in that case we block
        # the streams to wait until the buffer is free.
        wait_until_producer_ebb = app.producer.buffer.wait_until_ebb

        # This sensor method is called every time the buffer is full.
        on_topic_buffer_full = app.sensors.on_topic_buffer_full

        # callback called when the queue is under high pressure/
        # about to become full.
        def on_pressure_high() -> None:
            on_topic_buffer_full(tp)
            consumer_on_buffer_full(tp)

        # callback used when pressure drops.
        # added to Queue._pending_pressure_drop_callbacks
        # when the buffer is under high pressure/full.
        def on_pressure_drop() -> None:
            consumer_on_buffer_drop(tp)

        async def on_message(message: Message) -> None:
            # when a message is received we find all channels
            # that subscribe to this message
            await acquire_flow_control()
            await wait_until_producer_ebb()
            channels_n = len_(channels)
            if channels_n:
                # we increment the reference count for this message in bulk
                # immediately, so that nothing will get a chance to decref to
                # zero before we've had the chance to pass it to all channels
                message.incref(channels_n)
                event: Optional[EventT] = None
                event_keyid: Optional[Tuple[K, V]] = None

                # forward message to all channels subscribing to this topic

                # keep track of the number of channels we delivered to,
                # so that if a DecodeError is raised we can propagate
                # that error to the remaining channels.
                delivered: Set[_Topic] = set()
                full: typing.List[Tuple[EventT, _Topic]] = []
                try:
                    for chan in channels:
                        keyid = chan.key_type, chan.value_type
                        if event is None:
                            # first channel deserializes the payload:
                            event = await chan.decode(message, propagate=True)
                            event_keyid = keyid

                            queue = chan.queue
                            if queue.full():
                                full.append((event, chan))
                                continue
                            queue.put_nowait_enhanced(
                                event,
                                on_pressure_high=on_pressure_high,
                                on_pressure_drop=on_pressure_drop,
                            )
                        else:
                            # subsequent channels may have a different
                            # key/value type pair, meaning they all can
                            # deserialize the message in different ways

                            dest_event: EventT
                            if keyid == event_keyid:
                                # Reuse the event if it uses the same keypair:
                                dest_event = event
                            else:
                                dest_event = await chan.decode(message, propagate=True)
                            queue = chan.queue
                            if queue.full():
                                full.append((dest_event, chan))
                                continue
                            queue.put_nowait_enhanced(
                                dest_event,
                                on_pressure_high=on_pressure_high,
                                on_pressure_drop=on_pressure_drop,
                            )
                        delivered.add(chan)
                    if full:
                        for _, dest_chan in full:
                            on_topic_buffer_full(dest_chan)
                        await asyncio.wait(
                            [
                                dest_chan.put(dest_event)
                                for dest_event, dest_chan in full
                            ],
                            return_when=asyncio.ALL_COMPLETED,
                        )

                except KeyDecodeError as exc:
                    remaining = channels - delivered
                    message.ack(app.consumer, n=len(remaining))
                    for channel in remaining:
                        await channel.on_key_decode_error(exc, message)
                        delivered.add(channel)
                except ValueDecodeError as exc:
                    remaining = channels - delivered
                    message.ack(app.consumer, n=len(remaining))
                    for channel in remaining:
                        await channel.on_value_decode_error(exc, message)
                        delivered.add(channel)

        return on_message


class Conductor(ConductorT, Service):
    """Manages the channels that subscribe to topics.

    - Consumes messages from topic using a single consumer.
    - Forwards messages to all channels subscribing to a topic.
    """

    logger = logger

    #: Fast index to see if Topic is registered.
    _topics: MutableSet[TopicT]

    #: Map of (topic,partition) to set of channels that subscribe to that TP.
    _tp_index: MutableMapping[TP, MutableSet[TopicT]]

    #: Map str topic name to set of channels that subscribe
    #: to that topic.
    _topic_name_index: MutableMapping[str, MutableSet[TopicT]]

    #: For every TP assigned we compile a callback closure,
    #: and when we receive a message to a TP we look up which callback
    #: to call here.
    _tp_to_callback: MutableMapping[TP, ConsumerCallback]

    #: Whenever a change is made, i.e. a Topic is added/removed, we notify
    #: the background task responsible for resubscribing.
    _subscription_changed: Optional[asyncio.Event]

    _subscription_done: Optional[asyncio.Future]

    _acking_topics: Set[str]

    _compiler: ConductorCompiler

    #: We wait for 45 seconds after a resubscription request, to make
    #: sure any later requests are handled at the same time.
    _resubscribe_sleep_lock_seconds: float = 45.0

    def __init__(self, app: AppT, **kwargs: Any) -> None:
        Service.__init__(self, **kwargs)
        self.app = app
        self._topics = set()
        self._topic_name_index = defaultdict(set)
        self._tp_index = defaultdict(set)
        self._tp_to_callback = {}
        self._acking_topics = set()
        self._subscription_changed = None
        self._subscription_done = None
        self._compiler = ConductorCompiler()

        # This callback is called whenever the Consumer has
        # fetched a new record from Kafka.
        # We compile this down to a closure having variables already
        # localized as an optimization.
        self.on_message: ConsumerCallback
        self.on_message = self._compile_message_handler()

    async def commit(self, topics: TPorTopicSet) -> bool:
        """Commit offsets in topics."""
        return await self.app.consumer.commit(topics)

    def acks_enabled_for(self, topic: str) -> bool:
        """Return :const:`True` if acks are enabled for topic by name."""
        return topic in self._acking_topics

    def _compile_message_handler(self) -> ConsumerCallback:
        # This method localizes variables and attribute access
        # for better performance.  This is part of the inner loop
        # of a Faust worker, so tiny improvements here has big impact.

        get_callback_for_tp = self._tp_to_callback.__getitem__

        if self.app.client_only:

            async def on_message(message: Message) -> None:
                tp = TP(topic=message.topic, partition=0)
                return await get_callback_for_tp(tp)(message)

        else:

            async def on_message(message: Message) -> None:
                return await get_callback_for_tp(message.tp)(message)

        return on_message

    @Service.task
    async def _subscriber(self) -> None:  # pragma: no cover
        # the first time we start, we will wait two seconds
        # to give agents a chance to start up and register their
        # streams.  This way we won't have N subscription requests at the
        # start.
        if self.app.client_only or self.app.producer_only:
            self.log.info("Not waiting for agent/table startups...")
        else:
            self.log.info("Waiting for agents to start...")
            await self.app.agents.wait_until_agents_started()
            self.log.info("Waiting for tables to be registered...")
            await self.app.tables.wait_until_tables_registered()
        if not self.should_stop:
            # tell the consumer to subscribe to the topics.
            await self.app.consumer.subscribe(await self._update_indices())
            notify(self._subscription_done)

            # Now we wait for changes
            ev = self._subscription_changed = asyncio.Event(loop=self.loop)
        while not self.should_stop:
            # Wait for something to add/remove topics from subscription.
            await ev.wait()
            if self.app.rebalancing:
                # we do not want to perform a resubscribe if the application
                # is rebalancing.
                ev.clear()
            else:
                # The change could be in reaction to something like "all agents
                # restarting", in that case it would be bad if we resubscribe
                # over and over, so we wait for 45 seconds to make sure any
                # further subscription requests will happen during the same
                # rebalance.
                await self.sleep(self._resubscribe_sleep_lock_seconds)
                subscribed_topics = await self._update_indices()
                await self.app.consumer.subscribe(subscribed_topics)

            # clear the subscription_changed flag, so we can wait on it again.
            ev.clear()
            # wake-up anything waiting for the subscription to be done.
            notify(self._subscription_done)

    async def wait_for_subscriptions(self) -> None:
        """Wait for consumer to be subscribed."""
        if self._subscription_done is None:
            self._subscription_done = asyncio.Future(loop=self.loop)
        await self._subscription_done

    async def maybe_wait_for_subscriptions(self) -> None:
        if self._subscription_done is not None:
            await self._subscription_done

    async def _update_indices(self) -> Iterable[str]:
        self._topic_name_index.clear()
        self._tp_to_callback.clear()
        for channel in self._topics:
            if channel.internal:
                await channel.maybe_declare()
            for topic in channel.topics:
                if channel.acks:
                    self._acking_topics.add(topic)
                self._topic_name_index[topic].add(channel)

        return self._topic_name_index

    async def on_partitions_assigned(self, assigned: Set[TP]) -> None:
        """Call when cluster is rebalancing and partitions are assigned."""
        T = traced_from_parent_span()
        self._tp_index.clear()
        T(self._update_tp_index)(assigned)
        T(self._update_callback_map)()

    async def on_client_only_start(self) -> None:
        tp_index = self._tp_index
        for topic in self._topics:
            for subtopic in topic.topics:
                tp = TP(topic=subtopic, partition=0)
                tp_index[tp].add(topic)
        self._update_callback_map()

    def _update_tp_index(self, assigned: Set[TP]) -> None:
        assignmap = tp_set_to_map(assigned)
        tp_index = self._tp_index
        for topic in self._topics:
            if topic.active_partitions is not None:
                # Isolated Partitions: One agent per partition.
                if topic.active_partitions:
                    if assigned:
                        assert topic.active_partitions.issubset(assigned)
                    for tp in topic.active_partitions:
                        tp_index[tp].add(topic)
            else:
                # Default: One agent receives messages for all partitions.
                for subtopic in topic.topics:
                    for tp in assignmap[subtopic]:
                        tp_index[tp].add(topic)

    def _update_callback_map(self) -> None:
        self._tp_to_callback.update(
            (tp, self._build_handler(tp, cast(MutableSet[_Topic], channels)))
            for tp, channels in self._tp_index.items()
        )

    def _build_handler(self, tp: TP, channels: MutableSet[_Topic]) -> ConsumerCallback:
        if ConductorHandler is not None:  # pragma: no cover
            return ConductorHandler(self, tp, channels)
        else:
            return self._compiler.build(self, tp, channels)

    def clear(self) -> None:
        """Clear all subscriptions."""
        self._topics.clear()
        self._topic_name_index.clear()
        self._tp_index.clear()
        self._tp_to_callback.clear()
        self._acking_topics.clear()

    def __contains__(self, value: Any) -> bool:
        return value in self._topics

    def __iter__(self) -> Iterator[TopicT]:
        return iter(self._topics)

    def __len__(self) -> int:
        return len(self._topics)

    def __hash__(self) -> int:
        return object.__hash__(self)

    def add(self, topic: TopicT) -> None:
        """Register topic to be subscribed."""
        if topic not in self._topics:
            self._topics.add(topic)
            if self._topic_contain_unsubscribed_topics(topic):
                self._flag_changes()

    def _topic_contain_unsubscribed_topics(self, topic: TopicT) -> bool:
        index = self._topic_name_index
        return bool(index and any(t not in index for t in topic.topics))

    def discard(self, topic: Any) -> None:
        """Unregister topic from conductor."""
        self._topics.discard(topic)

    def _flag_changes(self) -> None:
        if self._subscription_changed is not None:
            self._subscription_changed.set()
        if self._subscription_done is None:
            self._subscription_done = asyncio.Future(loop=self.loop)

    @property
    def label(self) -> str:
        """Return label for use in logs."""
        return f"{type(self).__name__}({len(self._topics)})"

    @property
    def shortlabel(self) -> str:
        """Return short label for use in logs."""
        return type(self).__name__

    @property
    def acking_topics(self) -> Set[str]:
        return self._acking_topics



# cython: language_level=3
from asyncio import sleep
from time import monotonic

from mode.utils.futures import maybe_async, notify

from faust.exceptions import Skip
from faust.types import ChannelT, EventT


cdef class StreamIterator:

    cdef public:
        object stream
        object channel
        bint chan_is_channel
        object chan_queue
        object chan_queue_empty
        object chan_errors
        object chan_quick_get
        object chan_slow_get
        object processors
        object loop
        object on_merge
        object on_stream_event_in
        object on_stream_event_out
        object on_message_in
        object on_message_out
        object acking_topics
        object consumer
        object unacked
        object add_unacked
        object app
        object topics
        object acks_enabled_for
        object _skipped_value

    def __init__(self, object stream):
        self.stream = stream
        self.channel = self.stream.channel
        self.app = self.stream.app
        self.topics = self.app.topics
        self.acks_enabled_for = self.topics.acks_enabled_for
        self.loop = self.stream.loop
        self.on_merge = self.stream.on_merge
        self.on_stream_event_in = self.stream._on_stream_event_in
        self.on_stream_event_out = self.stream._on_stream_event_out
        self.on_message_in = self.stream._on_message_in
        self.on_message_out = self.stream._on_message_out
        self.acking_topics = stream.app.topics._acking_topics
        self.consumer = self.stream.app.consumer
        self.unacked = self.consumer._unacked_messages
        self.add_unacked = self.unacked.add
        self._skipped_value = self.stream._skipped_value

        if isinstance(self.channel, ChannelT):
            self.chan_is_channel = True
            self.chan_queue = self.channel.queue
            self.chan_queue_empty = self.chan_queue.empty
            self.chan_errors = self.chan_queue._errors
            self.chan_quick_get = self.chan_queue.get_nowait
        else:
            self.chan_is_channel = False
            self.chan_queue = None
            self.chan_queue_empty = None
            self.chan_errors = None
            self.chan_quick_get = None
        self.chan_slow_get = self.channel.__anext__
        self.processors = self.stream._processors

    async def next(self):
        cdef:
            object event
            object value
            object channel_value
            object stream
            object sensor_state
            bint enable_acks
        sensor_state = None
        stream = self.stream
        do_ack = stream.enable_acks

        value = None
        event = None

        while value is None and event is None:
            await sleep(0, loop=self.loop)
            need_slow_get, channel_value = self._try_get_quick_value()
            if need_slow_get:
                channel_value = await self.chan_slow_get()
            event, value, sensor_state = self._prepare_event(channel_value)

            try:
                for processor in self.processors:
                    value = await maybe_async(processor(value))
                value = await self.on_merge(value)
            except Skip:
                value = self._skipped_value
        return value, sensor_state

    cpdef object after(self, object event, object do_ack, object sensor_state):
        cdef:
            bint last_stream_to_ack
            int refcount
            object tp
            object offset
            object consumer
        consumer = self.consumer
        last_stream_to_ack = False
        if do_ack and event is not None:
            message = event.message
            if not message.acked:
                refcount = message.refcount
                refcount -= 1
                if refcount < 0:
                    refcount = 0
                message.refcount = refcount
                if not refcount:
                    message.acked = True
                    tp = message.tp
                    offset = message.offset
                    if self.acks_enabled_for(message.topic):
                        committed = consumer._committed_offset[tp]
                        try:
                            if committed is None or offset >= committed:
                                acked_index = consumer._acked_index[tp]
                                if offset not in acked_index:
                                    self.unacked.discard(message)
                                    acked_index.add(offset)
                                    acked_for_tp = consumer._acked[tp]
                                    acked_for_tp.append(offset)
                                    consumer._n_acked += 1
                                    last_stream_to_ack = True
                        finally:
                            notify(consumer._waiting_for_ack)
            tp = event.message.tp
            offset = event.message.offset
            self.on_stream_event_out(
                tp, offset, self.stream, event, sensor_state)
            if last_stream_to_ack:
                self.on_message_out(tp, offset, message)

    cdef object _prepare_event(self, object channel_value):
        cdef:
            object event
            object message
            object topic
            object tp
            object offset
            object consumer
            object stream_state
        stream_state = None
        if isinstance(channel_value, EventT):
            event = channel_value
            message = event.message
            topic = message.topic
            tp = message.tp
            offset = message.offset
            consumer = self.consumer

            if topic in self.acking_topics and not message.tracked:
                message.tracked = True
                self.add_unacked(message)
                self.on_message_in(tp, offset, message)

                stream_state = self.on_stream_event_in(
                    tp, offset, self.stream, event)
            self.stream._set_current_event(event)
            return (event, event.value, stream_state)
        else:
            self.stream._set_current_event(None)
            return None, channel_value, stream_state

    cdef object _try_get_quick_value(self):
        if self.chan_is_channel:
            if self.chan_errors:
                raise self.chan_errors.popleft()
            if self.chan_queue_empty:
                return (True, None)
            else:
                return self.chan_quick_get()
        return (True, None)



"""Monitor using Statsd."""
import typing
from typing import Any, Dict, Optional, cast

from mode.utils.objects import cached_property

from faust import web
from faust.exceptions import ImproperlyConfigured
from faust.types import (
    TP,
    AppT,
    CollectionT,
    EventT,
    Message,
    PendingMessage,
    RecordMetadata,
    StreamT,
)
from faust.types.assignor import PartitionAssignorT
from faust.types.transports import ConsumerT, ProducerT

from .monitor import Monitor, TPOffsetMapping

try:
    import statsd
except ImportError:  # pragma: no cover
    statsd = None

if typing.TYPE_CHECKING:  # pragma: no cover
    from statsd import StatsClient
else:

    class StatsClient:
        ...  # noqa


__all__ = ["StatsdMonitor"]


class StatsdMonitor(Monitor):
    """Statsd Faust Sensor.

    This sensor, records statistics to Statsd along with computing metrics
    for the stats server
    """

    host: str
    port: int
    prefix: str

    def __init__(
        self,
        host: str = "localhost",
        port: int = 8125,
        prefix: str = "faust-app",
        rate: float = 1.0,
        **kwargs: Any,
    ) -> None:
        self.host = host
        self.port = port
        self.prefix = prefix
        self.rate = rate
        if statsd is None:
            raise ImproperlyConfigured("StatsMonitor requires `pip install statsd`.")
        super().__init__(**kwargs)

    def _new_statsd_client(self) -> StatsClient:
        return statsd.StatsClient(host=self.host, port=self.port, prefix=self.prefix)

    def on_message_in(self, tp: TP, offset: int, message: Message) -> None:
        """Call before message is delegated to streams."""
        super().on_message_in(tp, offset, message)

        self.client.incr("messages_received", rate=self.rate)
        self.client.incr("messages_active", rate=self.rate)
        self.client.incr(f"topic.{tp.topic}.messages_received", rate=self.rate)
        self.client.gauge(f"read_offset.{tp.topic}.{tp.partition}", offset)

    def on_stream_event_in(
        self, tp: TP, offset: int, stream: StreamT, event: EventT
    ) -> Optional[Dict]:
        """Call when stream starts processing an event."""
        state = super().on_stream_event_in(tp, offset, stream, event)
        self.client.incr("events", rate=self.rate)
        self.client.incr(
            f"stream.{self._stream_label(stream)}.events",
            rate=self.rate,
        )
        self.client.incr("events_active", rate=self.rate)
        return state

    def _stream_label(self, stream: StreamT) -> str:
        return (
            self._normalize(
                stream.shortlabel.lstrip("Stream:"),
            )
            .strip("_")
            .lower()
        )

    def on_stream_event_out(
        self, tp: TP, offset: int, stream: StreamT, event: EventT, state: Dict = None
    ) -> None:
        """Call when stream is done processing an event."""
        super().on_stream_event_out(tp, offset, stream, event, state)
        self.client.decr("events_active", rate=self.rate)
        if state is not None:
            self.client.timing(
                "events_runtime",
                self.secs_to_ms(self.events_runtime[-1]),
                rate=self.rate,
            )

    def on_message_out(self, tp: TP, offset: int, message: Message) -> None:
        """Call when message is fully acknowledged and can be committed."""
        super().on_message_out(tp, offset, message)
        self.client.decr("messages_active", rate=self.rate)

    def on_table_get(self, table: CollectionT, key: Any) -> None:
        """Call when value in table is retrieved."""
        super().on_table_get(table, key)
        self.client.incr(f"table.{table.name}.keys_retrieved", rate=self.rate)

    def on_table_set(self, table: CollectionT, key: Any, value: Any) -> None:
        """Call when new value for key in table is set."""
        super().on_table_set(table, key, value)
        self.client.incr(f"table.{table.name}.keys_updated", rate=self.rate)

    def on_table_del(self, table: CollectionT, key: Any) -> None:
        """Call when key in a table is deleted."""
        super().on_table_del(table, key)
        self.client.incr(f"table.{table.name}.keys_deleted", rate=self.rate)

    def on_commit_completed(self, consumer: ConsumerT, state: Any) -> None:
        """Call when consumer commit offset operation completed."""
        super().on_commit_completed(consumer, state)
        self.client.timing(
            "commit_latency", self.ms_since(cast(float, state)), rate=self.rate
        )

    def on_send_initiated(
        self,
        producer: ProducerT,
        topic: str,
        message: PendingMessage,
        keysize: int,
        valsize: int,
    ) -> Any:
        """Call when message added to producer buffer."""
        self.client.incr(f"topic.{topic}.messages_sent", rate=self.rate)
        return super().on_send_initiated(producer, topic, message, keysize, valsize)

    def on_send_completed(
        self, producer: ProducerT, state: Any, metadata: RecordMetadata
    ) -> None:
        """Call when producer finished sending message."""
        super().on_send_completed(producer, state, metadata)
        self.client.incr("messages_sent", rate=self.rate)
        self.client.timing(
            "send_latency", self.ms_since(cast(float, state)), rate=self.rate
        )

    def on_send_error(
        self, producer: ProducerT, exc: BaseException, state: Any
    ) -> None:
        """Call when producer was unable to publish message."""
        super().on_send_error(producer, exc, state)
        self.client.incr("messages_sent_error", rate=self.rate)
        self.client.timing(
            "send_latency_for_error", self.ms_since(cast(float, state)), rate=self.rate
        )

    def on_assignment_error(
        self, assignor: PartitionAssignorT, state: Dict, exc: BaseException
    ) -> None:
        """Partition assignor did not complete assignor due to error."""
        super().on_assignment_error(assignor, state, exc)
        self.client.incr("assignments_error", rate=self.rate)
        self.client.timing(
            "assignment_latency", self.ms_since(state["time_start"]), rate=self.rate
        )

    def on_assignment_completed(
        self, assignor: PartitionAssignorT, state: Dict
    ) -> None:
        """Partition assignor completed assignment."""
        super().on_assignment_completed(assignor, state)
        self.client.incr("assignments_complete", rate=self.rate)
        self.client.timing(
            "assignment_latency", self.ms_since(state["time_start"]), rate=self.rate
        )

    def on_rebalance_start(self, app: AppT) -> Dict:
        """Cluster rebalance in progress."""
        state = super().on_rebalance_start(app)
        self.client.incr("rebalances", rate=self.rate)
        return state

    def on_rebalance_return(self, app: AppT, state: Dict) -> None:
        """Consumer replied assignment is done to broker."""
        super().on_rebalance_return(app, state)
        self.client.decr("rebalances", rate=self.rate)
        self.client.incr("rebalances_recovering", rate=self.rate)
        self.client.timing(
            "rebalance_return_latency",
            self.ms_since(state["time_return"]),
            rate=self.rate,
        )

    def on_rebalance_end(self, app: AppT, state: Dict) -> None:
        """Cluster rebalance fully completed (including recovery)."""
        super().on_rebalance_end(app, state)
        self.client.decr("rebalances_recovering", rate=self.rate)
        self.client.timing(
            "rebalance_end_latency", self.ms_since(state["time_end"]), rate=self.rate
        )

    def count(self, metric_name: str, count: int = 1) -> None:
        """Count metric by name."""
        super().count(metric_name, count=count)
        self.client.incr(metric_name, count=count, rate=self.rate)

    def on_tp_commit(self, tp_offsets: TPOffsetMapping) -> None:
        """Call when offset in topic partition is committed."""
        super().on_tp_commit(tp_offsets)
        for tp, offset in tp_offsets.items():
            metric_name = f"committed_offset.{tp.topic}.{tp.partition}"
            self.client.gauge(metric_name, offset)

    def track_tp_end_offset(self, tp: TP, offset: int) -> None:
        """Track new topic partition end offset for monitoring lags."""
        super().track_tp_end_offset(tp, offset)
        metric_name = f"end_offset.{tp.topic}.{tp.partition}"
        self.client.gauge(metric_name, offset)

    def on_web_request_end(
        self,
        app: AppT,
        request: web.Request,
        response: Optional[web.Response],
        state: Dict,
        *,
        view: web.View = None,
    ) -> None:
        """Web server finished working on request."""
        super().on_web_request_end(app, request, response, state, view=view)
        status_code = int(state["status_code"])
        self.client.incr(f"http_status_code.{status_code}", rate=self.rate)
        self.client.timing(
            "http_response_latency", self.ms_since(state["time_end"]), rate=self.rate
        )

    @cached_property
    def client(self) -> StatsClient:
        """Return statsd client."""
        return self._new_statsd_client()



"""LiveCheck - End-to-end testing of asynchronous systems."""
from .app import LiveCheck
from .case import Case
from .locals import current_test
from .runners import TestRunner
from .signals import Signal

__all__ = ["LiveCheck", "Case", "Signal", "TestRunner", "current_test"]



import logging
import os
import socket
import ssl
import typing
from datetime import timedelta, timezone, tzinfo
from pathlib import Path
from typing import (
    Any,
    Callable,
    ClassVar,
    Iterable,
    List,
    Mapping,
    Optional,
    Type,
    Union,
)
from uuid import uuid4

from mode import SupervisorStrategyT
from mode.utils.imports import SymbolArg, symbol_by_name
from mode.utils.logging import Severity
from mode.utils.times import Seconds, want_seconds
from yarl import URL

from faust.types._env import DATADIR, WEB_BIND, WEB_PORT, WEB_TRANSPORT
from faust.types.agents import AgentT
from faust.types.assignor import LeaderAssignorT, PartitionAssignorT
from faust.types.auth import CredentialsArg, CredentialsT
from faust.types.codecs import CodecArg
from faust.types.enums import ProcessingGuarantee
from faust.types.events import EventT
from faust.types.router import RouterT
from faust.types.sensors import SensorT
from faust.types.serializers import RegistryT, SchemaT
from faust.types.streams import StreamT
from faust.types.tables import GlobalTableT, TableManagerT, TableT
from faust.types.topics import TopicT
from faust.types.transports import PartitionerT, SchedulingStrategyT
from faust.types.web import HttpClientT, ResourceOptions

from . import base, params, sections
from .params import BrokerArg, URLArg

if typing.TYPE_CHECKING:
    from faust.types.worker import Worker as _WorkerT
else:

    class _WorkerT:
        ...  # noqa


# XXX mypy borks if we do `from faust import __version__`
faust_version: str = symbol_by_name("faust:__version__")

AutodiscoverArg = Union[
    bool,
    Iterable[str],
    Callable[[], Iterable[str]],
]


class Settings(base.SettingsRegistry):
    NODE_HOSTNAME: ClassVar[str] = socket.gethostname()
    DEFAULT_BROKER_URL: ClassVar[str] = "kafka://localhost:9092"

    _id: str
    _name: str
    _version: int

    #: Environment.
    #: Defaults to :data:`os.environ`.
    env: Mapping[str, str]

    def __init__(
        self,
        id: str,
        *,
        # Common settings:
        autodiscover: AutodiscoverArg = None,
        datadir: typing.Union[str, Path] = None,
        tabledir: typing.Union[str, Path] = None,
        debug: bool = None,
        env_prefix: str = None,
        id_format: str = None,
        origin: str = None,
        timezone: typing.Union[str, tzinfo] = None,
        version: int = None,
        # Agent settings:
        agent_supervisor: SymbolArg[Type[SupervisorStrategyT]] = None,
        # Broker settings:
        broker: BrokerArg = None,
        broker_consumer: BrokerArg = None,
        broker_producer: BrokerArg = None,
        broker_api_version: str = None,
        broker_check_crcs: bool = None,
        broker_client_id: str = None,
        broker_commit_every: int = None,
        broker_commit_interval: Seconds = None,
        broker_commit_livelock_soft_timeout: Seconds = None,
        broker_credentials: CredentialsArg = None,
        broker_heartbeat_interval: Seconds = None,
        broker_max_poll_interval: Seconds = None,
        broker_max_poll_records: int = None,
        broker_rebalance_timeout: Seconds = None,
        broker_request_timeout: Seconds = None,
        broker_session_timeout: Seconds = None,
        ssl_context: ssl.SSLContext = None,
        # Consumer settings:
        consumer_api_version: str = None,
        consumer_max_fetch_size: int = None,
        consumer_auto_offset_reset: str = None,
        consumer_group_instance_id: str = None,
        # Topic serialization settings:
        key_serializer: CodecArg = None,
        value_serializer: CodecArg = None,
        # Logging settings:
        logging_config: Mapping = None,
        loghandlers: List[logging.Handler] = None,
        # Producer settings:
        producer_acks: int = None,
        producer_api_version: str = None,
        producer_compression_type: str = None,
        producer_linger_ms: int = None,
        producer_max_batch_size: int = None,
        producer_max_request_size: int = None,
        producer_partitioner: SymbolArg[PartitionerT] = None,
        producer_request_timeout: Seconds = None,
        producer_threaded: bool = False,
        # RPC settings:
        reply_create_topic: bool = None,
        reply_expires: Seconds = None,
        reply_to: str = None,
        reply_to_prefix: str = None,
        # Stream settings:
        processing_guarantee: Union[str, ProcessingGuarantee] = None,
        stream_buffer_maxsize: int = None,
        stream_processing_timeout: Seconds = None,
        stream_publish_on_commit: bool = None,
        stream_recovery_delay: Seconds = None,
        stream_wait_empty: bool = None,
        # Table settings:
        store: URLArg = None,
        table_cleanup_interval: Seconds = None,
        table_key_index_size: int = None,
        table_standby_replicas: int = None,
        # Topic settings:
        topic_allow_declare: bool = None,
        topic_disable_leader: bool = None,
        topic_partitions: int = None,
        topic_replication_factor: int = None,
        # Web server settings:
        cache: URLArg = None,
        canonical_url: URLArg = None,
        web: URLArg = None,
        web_bind: str = None,
        web_cors_options: typing.Mapping[str, ResourceOptions] = None,
        web_enabled: bool = None,
        web_host: str = None,
        web_in_thread: bool = None,
        web_port: int = None,
        web_ssl_context: ssl.SSLContext = None,
        web_transport: URLArg = None,
        # Worker settings:
        worker_redirect_stdouts: bool = None,
        worker_redirect_stdouts_level: Severity = None,
        # Extension settings:
        Agent: SymbolArg[Type[AgentT]] = None,
        ConsumerScheduler: SymbolArg[Type[SchedulingStrategyT]] = None,
        Event: SymbolArg[Type[EventT]] = None,
        Schema: SymbolArg[Type[SchemaT]] = None,
        Stream: SymbolArg[Type[StreamT]] = None,
        Table: SymbolArg[Type[TableT]] = None,
        SetTable: SymbolArg[Type[TableT]] = None,
        GlobalTable: SymbolArg[Type[GlobalTableT]] = None,
        SetGlobalTable: SymbolArg[Type[GlobalTableT]] = None,
        TableManager: SymbolArg[Type[TableManagerT]] = None,
        Serializers: SymbolArg[Type[RegistryT]] = None,
        Worker: SymbolArg[Type[_WorkerT]] = None,
        PartitionAssignor: SymbolArg[Type[PartitionAssignorT]] = None,
        LeaderAssignor: SymbolArg[Type[LeaderAssignorT]] = None,
        Router: SymbolArg[Type[RouterT]] = None,
        Topic: SymbolArg[Type[TopicT]] = None,
        HttpClient: SymbolArg[Type[HttpClientT]] = None,
        Monitor: SymbolArg[Type[SensorT]] = None,
        # Deprecated settings:
        stream_ack_cancelled_tasks: bool = None,
        stream_ack_exceptions: bool = None,
        url: URLArg = None,
        **kwargs: Any,
    ) -> None:
        ...  # replaced by __init_subclass__ in BaseSettings

    def on_init(self, id: str, **kwargs: Any) -> None:
        # version is required for the id
        # and the id is required as a component in several default
        # setting values so we hack this in here to make sure
        # it's set.
        self._init_env_prefix(**kwargs)
        self._version = kwargs.get("version", 1)
        self.id = id

    def _init_env_prefix(
        self, env: Mapping[str, str] = None, env_prefix: str = None, **kwargs: Any
    ) -> None:
        if env is None:
            env = os.environ
        self.env = env
        env_name = self.SETTINGS["env_prefix"].env_name
        if env_name is not None:
            prefix_from_env = self.env.get(env_name)
            # prioritize environment
            if prefix_from_env is not None:
                self._env_prefix = prefix_from_env
            else:
                # then use provided argument
                if env_prefix is not None:
                    self._env_prefix = env_prefix

    def getenv(self, env_name: str) -> Any:
        if self._env_prefix:
            env_name = self._env_prefix.rstrip("_") + "_" + env_name
        return self.env.get(env_name)

    def relative_to_appdir(self, path: Path) -> Path:
        """Prepare app directory path.

        If path is absolute the path is returned as-is,
        but if path is relative it will be assumed to belong
        under the app directory.
        """
        return path if path.is_absolute() else self.appdir / path

    def data_directory_for_version(self, version: int) -> Path:
        """Return the directory path for data belonging to specific version."""
        return self.datadir / f"v{version}"

    def find_old_versiondirs(self) -> Iterable[Path]:
        for version in reversed(range(0, self.version)):
            path = self.data_directory_for_version(version)
            if path.is_dir():
                yield path

    @property
    def name(self) -> str:
        # name is a read-only property
        return self._name

    @property
    def id(self) -> str:
        return self._id

    @id.setter
    def id(self, name: str) -> None:
        self._name = name
        self._id = self._prepare_id(name)  # id is name+version

    def _prepare_id(self, id: str) -> str:
        if self.version > 1:
            return self.id_format.format(id=id, self=self)
        return id

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {self.id}>"

    @property
    def appdir(self) -> Path:
        return self.data_directory_for_version(self.version)

    # This is an example new setting having type ``str``
    @sections.Common.setting(
        params.Str,
        env_name="ENVIRONMENT_VARIABLE_NAME",
        version_removed="1.0",  # this disables the setting
    )
    def MY_SETTING(self) -> str:
        """My custom setting.

        To contribute new settings you only have to define a new
        setting decorated attribute here.

        Look at the other settings for examples.

        Remember that once you've added the setting you must
        also render the configuration reference:

        .. sourcecode:: console

            $ make configref
        """

    @sections.Common.setting(
        params.Param[AutodiscoverArg, AutodiscoverArg],
        default=False,
    )
    def autodiscover(self) -> AutodiscoverArg:
        """Automatic discovery of agents, tasks, timers, views and commands.

        Faust has an API to add different :mod:`asyncio` services and other
        user extensions, such as "Agents", HTTP web views,
        command-line commands, and timers to your Faust workers.
        These can be defined in any module, so to discover them at startup,
        the worker needs to traverse packages looking for them.

        .. warning::

            The autodiscovery functionality uses the :pypi:`Venusian` library
            to scan wanted packages for ``@app.agent``, ``@app.page``,
            ``@app.command``, ``@app.task`` and ``@app.timer`` decorators,
            but to do so, it's required to traverse the package path and import
            every module in it.

            Importing random modules like this can be dangerous so make sure
            you follow Python programming best practices. Do not start
            threads; perform network I/O; do test monkey-patching for mocks or
            similar, as a side effect of importing a module.  If you encounter
            a case such as this then please find a way to perform your
            action in a lazy manner.

        .. warning::

            If the above warning is something you cannot fix, or if it's out
            of your control, then please set ``autodiscover=False`` and make
            sure the worker imports all modules where your
            decorators are defined.

        The value for this argument can be:

        ``bool``
            If ``App(autodiscover=True)`` is set, the autodiscovery will
            scan the package name described in the ``origin`` attribute.

            The ``origin`` attribute is automatically set when you start
            a worker using the :program:`faust` command line program,
            for example:

            .. sourcecode:: console

                faust -A example.simple worker

            The :option:`-A <faust -A>`, option specifies the app, but you
            can also create a shortcut entry point by calling ``app.main()``:

            .. sourcecode:: python

                if __name__ == '__main__':
                    app.main()

            Then you can start the :program:`faust` program by executing for
            example ``python myscript.py worker --loglevel=INFO``, and it
            will use the correct application.

        ``Sequence[str]``
            The argument can also be a list of packages to scan::

                app = App(..., autodiscover=['proj_orders', 'proj_accounts'])

        ``Callable[[], Sequence[str]]``
            The argument can also be a function returning a list of packages
            to scan::

                def get_all_packages_to_scan():
                    return ['proj_orders', 'proj_accounts']

                app = App(..., autodiscover=get_all_packages_to_scan)

        ``False``
            If everything you need is in a self-contained module, or you
            import the stuff you need manually, just set ``autodiscover``
            to False and don't worry about it :-)

        .. admonition:: Django

            When using :pypi:`Django` and the :envvar:`DJANGO_SETTINGS_MODULE`
            environment variable is set, the Faust app will scan all packages
            found in the ``INSTALLED_APPS`` setting.

            If you're using Django you can use this to scan for
            agents/pages/commands in all packages
            defined in ``INSTALLED_APPS``.

            Faust will automatically detect that you're using Django
            and do the right thing if you do::

                app = App(..., autodiscover=True)

            It will find agents and other decorators in all of the
            reusable Django applications. If you want to manually control
            what packages are traversed, then provide a list::

                app = App(..., autodiscover=['package1', 'package2'])

            or if you want exactly :const:`None` packages to be traversed,
            then provide a False:

                app = App(.., autodiscover=False)

            which is the default, so you can simply omit the argument.

        .. tip::

            For manual control over autodiscovery, you can also call the
            :meth:`@discover` method manually.
        """

    @sections.Common.setting(
        params.Path,
        env_name="APP_DATADIR",
        default=DATADIR,
        related_cli_options={"faust": "--datadir"},
    )
    def datadir(self, path: Path) -> Path:
        """Application data directory.

        The directory in which this instance stores the data used by
        local tables, etc.

        .. seealso::

            - The data directory can also be set using the
              :option:`faust --datadir` option, from the command-line,
              so there is usually no reason to provide a default value
              when creating the app.
        """

    @datadir.on_get_value  # type: ignore
    def _prepare_datadir(self, path: Path) -> Path:
        # allow expanding variables in path
        return Path(str(path).format(conf=self))

    @sections.Common.setting(
        params.Path,
        #: This path will be treated as relative to datadir,
        #: unless the provided poth is absolute.
        default="tables",
        env_name="APP_TABLEDIR",
    )
    def tabledir(self) -> Path:
        """Application table data directory.

        The directory in which this instance stores local table data.
        Usually you will want to configure the :setting:`datadir` setting,
        but if you want to store tables separately you can configure this one.

        If the path provided is relative (it has no leading slash), then the
        path will be considered to be relative to the :setting:`datadir`
        setting.
        """

    @tabledir.on_get_value  # type: ignore
    def _prepare_tabledir(self, path: Path) -> Path:
        return self.relative_to_appdir(path)

    @sections.Common.setting(
        params.Bool,
        env_name="APP_DEBUG",
        default=False,
        related_cli_options={"faust": "--debug"},
    )
    def debug(self) -> bool:
        """Use in development to expose sensor information endpoint.


        .. tip::

            If you want to enable the sensor statistics endpoint in production,
            without enabling the :setting:`debug` setting, you can do so
            by adding the following code:

            .. sourcecode:: python

                app.web.blueprints.add(
                    '/stats/', 'faust.web.apps.stats:blueprint')
        """

    @sections.Common.setting(
        params.Str,
        env_name="APP_ENV_PREFIX",
        version_introduced="1.11",
        default=None,
        ignore_default=True,
    )
    def env_prefix(self) -> str:
        """Environment variable prefix.

        When configuring Faust by environent variables,
        this adds a common prefix to all Faust environment value names.
        """

    @sections.Common.setting(
        params.Str,
        env_name="APP_ID_FORMAT",
        default="{id}-v{self.version}",
    )
    def id_format(self) -> str:
        """Application ID format template.

        The format string used to generate the final :setting:`id` value
        by combining it with the :setting:`version` parameter.
        """

    @sections.Common.setting(
        params.Str,
        default=None,
    )
    def origin(self) -> str:
        """The reverse path used to find the app.

        For example if the app is located in::

            from myproj.app import app

        Then the ``origin`` should be ``"myproj.app"``.

        The :program:`faust worker` program will try to automatically set
        the origin, but if you are having problems with auto generated names
        then you can set origin manually.
        """

    @sections.Common.setting(
        params.Timezone,
        version_introduced="1.4",
        env_name="TIMEZONE",
        default=timezone.utc,
    )
    def timezone(self) -> tzinfo:
        """Project timezone.

        The timezone used for date-related functionality such as cronjobs.
        """

    @sections.Common.setting(
        params.Int,
        env_name="APP_VERSION",
        default=1,
        min_value=1,
    )
    def version(self) -> int:
        """App version.

        Version of the app, that when changed will create a new isolated
        instance of the application. The first version is 1,
        the second version is 2, and so on.

        .. admonition:: Source topics will not be affected by a version change.

            Faust applications will use two kinds of topics: source topics, and
            internally managed topics. The source topics are declared by the
            producer, and we do not have the opportunity to modify any
            configuration settings, like number of partitions for a source
            topic; we may only consume from them. To mark a topic as internal,
            use: ``app.topic(..., internal=True)``.
        """

    @sections.Agent.setting(
        params.Symbol(Type[SupervisorStrategyT]),
        env_name="AGENT_SUPERVISOR",
        default="mode.OneForOneSupervisor",
    )
    def agent_supervisor(self) -> Type[SupervisorStrategyT]:
        """Default agent supervisor type.

        An agent may start multiple instances (actors) when
        the concurrency setting is higher than one (e.g.
        ``@app.agent(concurrency=2)``).

        Multiple instances of the same agent are considered to be in the same
        supervisor group.

        The default supervisor is the :class:`mode.OneForOneSupervisor`:
        if an instance in the group crashes, we restart that instance only.

        These are the supervisors supported:

        + :class:`mode.OneForOneSupervisor`

            If an instance in the group crashes we restart only that instance.

        + :class:`mode.OneForAllSupervisor`

            If an instance in the group crashes we restart the whole group.

        + :class:`mode.CrashingSupervisor`

            If an instance in the group crashes we stop the whole application,
            and exit so that the Operating System supervisor can restart us.

        + :class:`mode.ForfeitOneForOneSupervisor`

            If an instance in the group crashes we give up on that instance
            and never restart it again (until the program is restarted).

        + :class:`mode.ForfeitOneForAllSupervisor`

            If an instance in the group crashes we stop all instances
            in the group and never restarted them again (until the program is
            restarted).
        """

    @sections.Common.setting(
        params.Seconds,
        env_name="BLOCKING_TIMEOUT",
        default=None,
        related_cli_options={"faust": "--blocking-timeout"},
    )
    def blocking_timeout(self) -> Optional[float]:
        """Blocking timeout (in seconds).

        When specified the worker will start a periodic signal based
        timer that only triggers when the loop has been blocked
        for a time exceeding this timeout.

        This is the most safe way to detect blocking, but could have
        adverse effects on libraries that do not automatically
        retry interrupted system calls.

        Python itself does retry all interrupted system calls
        since version 3.5 (see :pep:`475`), but this might not
        be the case with C extensions added to the worker by the user.

        The blocking detector is a background thread
        that periodically wakes up to either arm a timer, or cancel
        an already armed timer. In pseudocode:

        .. sourcecode:: python

            while True:
                # cancel previous alarm and arm new alarm
                signal.signal(signal.SIGALRM, on_alarm)
                signal.setitimer(signal.ITIMER_REAL, blocking_timeout)
                # sleep to wakeup just before the timeout
                await asyncio.sleep(blocking_timeout * 0.96)

            def on_alarm(signum, frame):
                logger.warning('Blocking detected: ...')

        If the sleep does not wake up in time the alarm signal
        will be sent to the process and a traceback will be logged.
        """

    @sections.Common.setting(
        params.BrokerList,
        env_name="BROKER_URL",
    )
    def broker(self) -> List[URL]:
        """Broker URL, or a list of alternative broker URLs.

        Faust needs the URL of a "transport" to send and receive messages.

        Currently, the only supported production transport is ``kafka://``.
        This uses the :pypi:`aiokafka` client under the hood, for consuming and
        producing messages.

        You can specify multiple hosts at the same time by separating them
        using the semi-comma:

        .. sourcecode:: text

            kafka://kafka1.example.com:9092;kafka2.example.com:9092

        Which in actual code looks like this:

        .. sourcecode:: python

            BROKERS = 'kafka://kafka1.example.com:9092;kafka2.example.com:9092'
            app = faust.App(
                'id',
                broker=BROKERS,
            )

        You can also pass a list of URLs:

        .. sourcecode:: python

            app = faust.App(
                'id',
                broker=['kafka://kafka1.example.com:9092',
                        'kafka://kafka2.example.com:9092'],
            )

        .. seealso::

            You can configure the transport used for consuming and producing
            separately, by setting the :setting:`broker_consumer` and
            :setting:`broker_producer` settings.

            This setting is used as the default.

        **Available Transports**

        - ``kafka://``

            Alias to ``aiokafka://``

        - ``aiokafka://``

            The recommended transport using the :pypi:`aiokafka` client.

            Limitations: None
        """

    @broker.on_set_default  # type: ignore
    def _prepare_broker(self) -> BrokerArg:
        return self._url or self.DEFAULT_BROKER_URL

    @sections.Broker.setting(
        params.BrokerList,
        version_introduced="1.7",
        env_name="BROKER_CONSUMER_URL",
        default_alias="broker",
    )
    def broker_consumer(self) -> List[URL]:
        """Consumer broker URL.

        You can use this setting to configure the transport used for
        producing and consuming separately.

        If not set the value found in :setting:`broker` will be used.
        """

    @sections.Broker.setting(
        params.BrokerList,
        version_introduced="1.7",
        env_name="BROKER_PRODUCER_URL",
        default_alias="broker",
    )
    def broker_producer(self) -> List[URL]:
        """Producer broker URL.

        You can use this setting to configure the transport used for
        producing and consuming separately.

        If not set the value found in :setting:`broker` will be used.
        """

    @sections.Broker.setting(
        params.Str,
        version_introduced="1.10",
        env_name="BROKER_API_VERSION",
        #: Default broker API version.
        #: Used as default for
        #:     + :setting:`broker_api_version`,
        #:     + :setting:`consumer_api_version`,
        #:     + :setting:`producer_api_version',
        default="auto",
    )
    def broker_api_version(self) -> str:
        """Broker API version,.

        This setting is also the default for :setting:`consumer_api_version`,
        and :setting:`producer_api_version`.

        Negotiate producer protocol version.

        The default value - "auto" means use the latest version supported
        by both client and server.

        Any other version set means you are requesting a specific version of
        the protocol.

        Example Kafka uses:

        **Disable sending headers for all messages produced**

        Kafka headers support was added in Kafka 0.11, so you can specify
        ``broker_api_version="0.10"`` to remove the headers from messages.
        """

    @sections.Broker.setting(
        params.Bool,
        env_name="BROKER_CHECK_CRCS",
        default=True,
    )
    def broker_check_crcs(self) -> bool:
        """Broker CRC check.

        Automatically check the CRC32 of the records consumed.
        """

    @sections.Broker.setting(
        params.Str,
        env_name="BROKER_CLIENT_ID",
        default=f"faust-{faust_version}",
    )
    def broker_client_id(self) -> str:
        """Broker client ID.

        There is rarely any reason to configure this setting.

        The client id is used to identify the software used, and is not usually
        configured by the user.
        """

    @sections.Broker.setting(
        params.UnsignedInt,
        env_name="BROKER_COMMIT_EVERY",
        default=10_000,
    )
    def broker_commit_every(self) -> int:
        """Broker commit message frequency.

        Commit offset every n messages.

        See also :setting:`broker_commit_interval`, which is how frequently
        we commit on a timer when there are few messages being received.
        """

    @sections.Broker.setting(
        params.Seconds,
        env_name="BROKER_COMMIT_INTERVAL",
        default=2.8,
    )
    def broker_commit_interval(self) -> float:
        """Broker commit time frequency.

        How often we commit messages that have been
        fully processed (:term:`acked`).
        """

    @sections.Broker.setting(
        params.Seconds,
        env_name="BROKER_COMMIT_LIVELOCK_SOFT_TIMEOUT",
        default=want_seconds(timedelta(minutes=5)),
    )
    def broker_commit_livelock_soft_timeout(self) -> float:
        """Commit livelock timeout.

        How long time it takes before we warn that the Kafka commit offset has
        not advanced (only when processing messages).
        """

    @sections.Common.setting(
        params.Credentials,
        version_introduced="1.5",
        env_name="BROKER_CREDENTIALS",
        default=None,
    )
    def broker_credentials(self) -> CredentialsT:
        """Broker authentication mechanism.

        Specify the authentication mechanism to use when connecting to
        the broker.

        The default is to not use any authentication.

        SASL Authentication
            You can enable SASL authentication via plain text:

            .. sourcecode:: python

                app = faust.App(
                    broker_credentials=faust.SASLCredentials(
                        username='x',
                        password='y',
                    ))

            .. warning::

                Do not use literal strings when specifying passwords in
                production, as they can remain visible in stack traces.

                Instead the best practice is to get the password from
                a configuration file, or from the environment:

                .. sourcecode:: python

                    BROKER_USERNAME = os.environ.get('BROKER_USERNAME')
                    BROKER_PASSWORD = os.environ.get('BROKER_PASSWORD')

                    app = faust.App(
                        broker_credentials=faust.SASLCredentials(
                            username=BROKER_USERNAME,
                            password=BROKER_PASSWORD,
                        ))

        GSSAPI Authentication
            GSSAPI authentication over plain text:

            .. sourcecode:: python

                app = faust.App(
                    broker_credentials=faust.GSSAPICredentials(
                        kerberos_service_name='faust',
                        kerberos_domain_name='example.com',
                    ),
                )

            GSSAPI authentication over SSL:

            .. sourcecode:: python

                import ssl
                ssl_context = ssl.create_default_context(
                    purpose=ssl.Purpose.SERVER_AUTH, cafile='ca.pem')
                ssl_context.load_cert_chain(
                    'client.cert', keyfile='client.key')

                app = faust.App(
                    broker_credentials=faust.GSSAPICredentials(
                        kerberos_service_name='faust',
                        kerberos_domain_name='example.com',
                        ssl_context=ssl_context,
                    ),
                )

        SSL Authentication
            Provide an SSL context for the Kafka broker connections.

            This allows Faust to use a secure SSL/TLS connection for the
            Kafka connections and enabling certificate-based authentication.

            .. sourcecode:: python

                import ssl

                ssl_context = ssl.create_default_context(
                    purpose=ssl.Purpose.SERVER_AUTH, cafile='ca.pem')
                ssl_context.load_cert_chain(
                    'client.cert', keyfile='client.key')
                app = faust.App(..., broker_credentials=ssl_context)
        """

    @sections.Broker.setting(
        params.Seconds,
        version_introduced="1.0.11",
        env_name="BROKER_HEARTBEAT_INTERVAL",
        default=3.0,
    )
    def broker_heartbeat_interval(self) -> float:
        """Broker heartbeat interval.

        How often we send heartbeats to the broker, and also how often
        we expect to receive heartbeats from the broker.

        If any of these time out, you should increase this setting.
        """

    @sections.Broker.setting(
        params.Seconds,
        version_introduced="1.7",
        env_name="BROKER_MAX_POLL_INTERVAL",
        default=1000.0,
    )
    def broker_max_poll_interval(self) -> float:
        """Broker max poll interval.

        The maximum allowed time (in seconds) between calls to consume messages
        If this interval is exceeded the consumer
        is considered failed and the group will rebalance in order to reassign
        the partitions to another consumer group member. If API methods block
        waiting for messages, that time does not count against this timeout.

        See `KIP-62`_ for technical details.

        .. _`KIP-62`:
            https://cwiki.apache.org/confluence/display/KAFKA/KIP-62%3A+Allow+consumer+to+send+heartbeats+from+a+background+thread
        """

    @sections.Broker.setting(
        params.UnsignedInt,
        version_introduced="1.4",
        env_name="BROKER_MAX_POLL_RECORDS",
        default=None,
        allow_none=True,
    )
    def broker_max_poll_records(self) -> Optional[int]:
        """Broker max poll records.

        The maximum number of records returned in a single call to ``poll()``.
        If you find that your application needs more time to process
        messages you may want to adjust :setting:`broker_max_poll_records`
        to tune the number of records that must be handled on every
        loop iteration.
        """

    @sections.Broker.setting(
        params.Seconds,
        version_introduced="1.10",
        env_name="BROKER_REBALANCE_TIMEOUT",
        default=60.0,
    )
    def broker_rebalance_timeout(self) -> float:
        """Broker rebalance timeout.

        How long to wait for a node to finish rebalancing before the broker
        will consider it dysfunctional and remove it from the cluster.

        Increase this if you experience the cluster being in a state of
        constantly rebalancing, but make sure you also increase the
        :setting:`broker_heartbeat_interval` at the same time.

        .. note::

            The session timeout must not be greater than the
            :setting:`broker_request_timeout`.
        """

    @sections.Broker.setting(
        params.Seconds,
        version_introduced="1.4",
        env_name="BROKER_REQUEST_TIMEOUT",
        default=90.0,
    )
    def broker_request_timeout(self) -> float:
        """Kafka client request timeout.

        .. note::

            The request timeout must not be less than the
            :setting:`broker_session_timeout`.
        """

    @sections.Broker.setting(
        params.Seconds,
        version_introduced="1.0.11",
        env_name="BROKER_SESSION_TIMEOUT",
        default=60.0,
    )
    def broker_session_timeout(self) -> float:
        """Broker session timeout.

        How long to wait for a node to finish rebalancing before the broker
        will consider it dysfunctional and remove it from the cluster.

        Increase this if you experience the cluster being in a state of
        constantly rebalancing, but make sure you also increase the
        :setting:`broker_heartbeat_interval` at the same time.

        .. note::

            The session timeout must not be greater than the
            :setting:`broker_request_timeout`.
        """

    @sections.Common.setting(
        params.SSLContext,
        default=None,
    )
    def ssl_context(self) -> ssl.SSLContext:
        """SSL configuration.

        See :setting:`credentials`.
        """

    @sections.Consumer.setting(
        params.Str,
        version_introduced="1.10",
        env_name="CONSUMER_API_VERSION",
        default_alias="broker_api_version",
    )
    def consumer_api_version(self) -> str:
        """Consumer API version.

        Configures the broker API version to use for consumers.
        See :setting:`broker_api_version` for more information.
        """

    @sections.Consumer.setting(
        params.UnsignedInt,
        version_introduced="1.4",
        env_name="CONSUMER_MAX_FETCH_SIZE",
        default=1024 ** 2,
    )
    def consumer_max_fetch_size(self) -> int:
        """Consumer max fetch size.

        The maximum amount of data per-partition the server will return.
        This size must be at least as large as the maximum message size.

        Note: This is PER PARTITION, so a limit of 1Mb when your
        workers consume from 10 topics having 100 partitions each,
        means a fetch request can be up to a gigabyte (10 * 100 * 1Mb),
        This limit being too generous may cause
        rebalancing issues: if the amount of time required
        to flush pending data stuck in socket buffers exceed
        the rebalancing timeout.

        You must keep this limit low enough to account
        for many partitions being assigned to a single node.
        """

    @sections.Consumer.setting(
        params.Str,
        version_introduced="1.5",
        env_name="CONSUMER_AUTO_OFFSET_RESET",
        default="earliest",
    )
    def consumer_auto_offset_reset(self) -> str:
        """Consumer auto offset reset.

        Where the consumer should start reading messages from when
        there is no initial offset, or the stored offset no longer
        exists, e.g. when starting a new consumer for the first time.

        Options include 'earliest', 'latest', 'none'.
        """

    @sections.Consumer.setting(
        params.Str,
        version_introduced="2.1",
        env_name="CONSUMER_GROUP_INSTANCE_ID",
        default=None,
    )
    def consumer_group_instance_id(self) -> str:
        """Consumer group instance id.

        The group_instance_id for static partition assignment.

        If not set, default assignment strategy is used. Otherwise,
        each consumer instance has to have a unique id.
        """

    @sections.Serialization.setting(
        params.Codec,
        env_name="APP_KEY_SERIALIZER",
        default="raw",
    )
    def key_serializer(self) -> CodecArg:
        """Default key serializer.

        Serializer used for keys by default when no serializer
        is specified, or a model is not being used.

        This can be the name of a serializer/codec, or an actual
        :class:`faust.serializers.codecs.Codec` instance.

        .. seealso::

            - The :ref:`codecs` section in the model guide -- for
              more information about codecs.
        """

    @sections.Serialization.setting(
        params.Codec,
        env_name="APP_VALUE_SERIALIZER",
        default="json",
    )
    def value_serializer(self) -> CodecArg:
        """Default value serializer.

        Serializer used for values by default when no serializer
        is specified, or a model is not being used.

        This can be string, the name of a serializer/codec, or an actual
        :class:`faust.serializers.codecs.Codec` instance.

        .. seealso::

            - The :ref:`codecs` section in the model guide -- for
              more information about codecs.
        """

    @sections.Common.setting(
        params.Dict[Any],
        version_introduced="1.5",
    )
    def logging_config(self) -> Mapping[str, Any]:
        """Logging dictionary configuration.

        Optional dictionary for logging configuration, as supported
        by :func:`logging.config.dictConfig`.
        """

    @sections.Common.setting(
        params.LogHandlers,
    )
    def loghandlers(self) -> List[logging.Handler]:
        """List of custom logging handlers.

        Specify a list of custom log handlers to use in worker instances.
        """

    @sections.Producer.setting(
        params.Int,
        env_name="PRODUCER_ACKS",
        default=-1,
        number_aliases={"all": -1},
    )
    def producer_acks(self) -> int:
        """Producer Acks.

        The number of acknowledgments the producer requires the leader to have
        received before considering a request complete. This controls the
        durability of records that are sent. The following settings are common:

        * ``0``: Producer will not wait for any acknowledgment from
                 the server at all. The message will immediately be
                 considered sent (Not recommended).
        * ``1``: The broker leader will write the record to its local
                 log but will respond without awaiting full acknowledgment
                 from all followers. In this case should the leader fail
                 immediately after acknowledging the record but before the
                 followers have replicated it then the record will be lost.
        * ``-1``: The broker leader will wait for the full set of in-sync
                  replicas to acknowledge the record. This guarantees that
                  the record will not be lost as long as at least one in-sync
                  replica remains alive. This is the strongest
                  available guarantee.
        """

    @sections.Producer.setting(
        params.Str,
        version_introduced="1.5.3",
        env_name="PRODUCER_API_VERSION",
        default_alias="broker_api_version",
    )
    def producer_api_version(self) -> str:
        """Producer API version.

        Configures the broker API version to use for producers.
        See :setting:`broker_api_version` for more information.
        """

    @sections.Producer.setting(
        params.Str,
        env_name="PRODUCER_COMPRESSION_TYPE",
        default=None,
    )
    def producer_compression_type(self) -> str:
        """Producer compression type.

        The compression type for all data generated by the producer.
        Valid values are `gzip`, `snappy`, `lz4`, or :const:`None`.
        """

    @sections.Producer.setting(
        params.Seconds,
        env_name="PRODUCER_LINGER",
    )
    def producer_linger(self) -> Optional[float]:
        """Producer batch linger configuration.

        Minimum time to batch before sending out messages from the producer.

        Should rarely have to change this.
        """

    @producer_linger.on_set_default  # type: ignore
    def _prepare_producer_linger(self) -> float:
        return float(self._producer_linger_ms) / 1000.0

    @sections.Producer.setting(
        params.UnsignedInt,
        env_name="PRODUCER_MAX_BATCH_SIZE",
        default=16384,
    )
    def producer_max_batch_size(self) -> int:
        """Producer max batch size.

        Max size of each producer batch, in bytes.
        """

    @sections.Producer.setting(
        params.UnsignedInt,
        env_name="PRODUCER_MAX_REQUEST_SIZE",
        default=1_000_000,
    )
    def producer_max_request_size(self) -> int:
        """Producer maximum request size.

        Maximum size of a request in bytes in the producer.

        Should rarely have to change this.
        """

    @sections.Producer.setting(
        params._Symbol[PartitionerT, Optional[PartitionerT]],
        version_introduced="1.2",
        default=None,
    )
    def producer_partitioner(self) -> Optional[PartitionerT]:
        """Producer partitioning strategy.

        The Kafka producer can be configured with a custom partitioner
        to change how keys are partitioned when producing to topics.

        The default partitioner for Kafka is implemented as follows,
        and can be used as a template for your own partitioner:

        .. sourcecode:: python

            import random
            from typing import List
            from kafka.partitioner.hashed import murmur2

            def partition(key: bytes,
                        all_partitions: List[int],
                        available: List[int]) -> int:
                '''Default partitioner.

                Hashes key to partition using murmur2 hashing
                (from java client) If key is None, selects partition
                randomly from available, or from all partitions if none
                are currently available

                Arguments:
                    key: partitioning key
                    all_partitions: list of all partitions sorted by
                                    partition ID.
                    available: list of available partitions
                               in no particular order
                Returns:
                    int: one of the values from ``all_partitions``
                         or ``available``.
                '''
                if key is None:
                    source = available if available else all_paritions
                    return random.choice(source)
                index: int = murmur2(key)
                index &= 0x7fffffff
                index %= len(all_partitions)
                return all_partitions[index]
        """

    @sections.Producer.setting(
        params.Seconds,
        version_introduced="1.4",
        env_name="PRODUCER_REQUEST_TIMEOUT",
        default=1200.0,  # 20 minutes
    )
    def producer_request_timeout(self) -> float:
        """Producer request timeout.

        Timeout for producer operations.
        This is set high by default, as this is also the time when
        producer batches expire and will no longer be retried.
        """

    @sections.Producer.setting(
        params.Bool,
        version_introduced="0.4.5",
        env_name="PRODUCER_THREADED",
        default=False,
    )
    def producer_threaded(self) -> bool:
        """Thread separate producer for send_soon.

        If True, spin up a different producer in a different thread
        to be used for messages buffered up for producing via
        send_soon function.
        """

    @sections.Stream.setting(
        params.Bool,
        version_introduced="0.4.7",
        env_name="RECOVERY_CONSISTENCY_CHECK",
        default=True,
    )
    def recovery_consistency_check(self) -> bool:
        """Check Kafka and local offsets for consistency.

        If True, assert that Kafka highwater offsets >= local offset
        in the rocksdb state storee
        """

    @sections.Stream.setting(
        params.Bool,
        version_introduced="0.6.0",
        env_name="STORE_CHECK_EXISTS",
        default=True,
    )
    def store_check_exists(self) -> bool:
        """Execute exists on the underlying store.

        If True, executes exists on the underlying store. If False
        client has to catch KeyError
        """

    @sections.RPC.setting(
        params.Bool,
        env_name="APP_REPLY_CREATE_TOPIC",
        default=False,
    )
    def reply_create_topic(self) -> bool:
        """Automatically create reply topics.

        Set this to :const:`True` if you plan on using the RPC with agents.

        This will create the internal topic used for RPC replies on that
        instance at startup.
        """

    @sections.RPC.setting(
        params.Seconds,
        env_name="APP_REPLY_EXPIRES",
        default=want_seconds(timedelta(days=1)),
    )
    def reply_expires(self) -> float:
        """RPC reply expiry time in seconds.

        The expiry time (in seconds :class:`float`,
        or :class:`~datetime.timedelta`), for how long replies will stay
        in the instances local reply topic before being removed.
        """

    @sections.RPC.setting(
        params.Str,
    )
    def reply_to(self) -> str:
        """Reply to address.

        The name of the reply topic used by this instance.
        If not set one will be automatically generated when the app
        is created.
        """

    @reply_to.on_set_default  # type: ignore
    def _prepare_reply_to_default(self) -> str:
        return f"{self.reply_to_prefix}{uuid4()}"

    @sections.RPC.setting(
        params.Str,
        env_name="APP_REPLY_TO_PREFIX",
        default="f-reply-",
    )
    def reply_to_prefix(self) -> str:
        """Reply address topic name prefix.

        The prefix used when generating reply topic names.
        """

    @sections.Common.setting(
        params.Enum(ProcessingGuarantee),
        version_introduced="1.5",
        env_name="PROCESSING_GUARANTEE",
        default=ProcessingGuarantee.AT_LEAST_ONCE,
    )
    def processing_guarantee(self) -> ProcessingGuarantee:
        """The processing guarantee that should be used.

        Possible values are "at_least_once" (default) and "exactly_once".

        Note that if exactly-once processing is enabled consumers are
        configured with ``isolation.level="read_committed"`` and producers
        are configured with ``retries=Integer.MAX_VALUE`` and
        ``enable.idempotence=true`` per default.

        Note that by default exactly-once processing requires a cluster of
        at least three brokers what is the recommended setting for production.
        For development you can change this, by adjusting broker setting
        ``transaction.state.log.replication.factor`` to the number of brokers
        you want to use.
        """

    @sections.Stream.setting(
        params.UnsignedInt,
        env_name="STREAM_BUFFER_MAXSIZE",
        default=4096,
    )
    def stream_buffer_maxsize(self) -> int:
        """Stream buffer maximum size.

        This setting control back pressure to streams and agents reading
        from streams.

        If set to 4096 (default) this means that an agent can only keep
        at most 4096 unprocessed items in the stream buffer.

        Essentially this will limit the number of messages a stream
        can "prefetch".

        Higher numbers gives better throughput, but do note that if
        your agent sends messages or update tables (which
        sends changelog messages).

        This means that if the buffer size is large, the
        :setting:`broker_commit_interval` or :setting:`broker_commit_every`
        settings must be set to commit frequently, avoiding back pressure
        from building up.

        A buffer size of 131_072 may let you process over 30,000 events
        a second as a baseline, but be careful with a buffer size that large
        when you also send messages or update tables.
        """

    @sections.Stream.setting(
        params.Seconds,
        version_introduced="1.10",
        env_name="STREAM_PROCESSING_TIMEOUT",
        default=5 * 60.0,
    )
    def stream_processing_timeout(self) -> float:
        """Stream processing timeout.

        Timeout (in seconds) for processing events in the stream.
        If processing of a single event exceeds this time we log an error,
        but do not stop processing.

        If you are seeing a warning like this you should either

        1) increase this timeout to allow agents to spend more time
            on a single event, or
        2) add a timeout to the operation in the agent, so stream processing
            always completes before the timeout.

        The latter is preferred for network operations such as web requests.
        If a network service you depend on is temporarily offline you should
        consider doing retries (send to separate topic):

        .. sourcecode:: python

            main_topic = app.topic('main')
            deadletter_topic = app.topic('main_deadletter')

            async def send_request(value, timeout: float = None) -> None:
                await app.http_client.get('http://foo.com', timeout=timeout)

            @app.agent(main_topic)
            async def main(stream):
                async for value in stream:
                try:
                    await send_request(value, timeout=5)
                except asyncio.TimeoutError:
                    await deadletter_topic.send(value)

            @app.agent(deadletter_topic)
                async def main_deadletter(stream):
                    async for value in stream:
                    # wait for 30 seconds before retrying.
                    await stream.sleep(30)
                    await send_request(value)
        """

    @sections.Stream.setting(
        params.Bool,
        default=False,
    )
    def stream_publish_on_commit(self) -> bool:
        """Stream delay producing until commit time.

        If enabled we buffer up sending messages until the
        source topic offset related to that processing is committed.
        This means when we do commit, we may have buffered up a LOT of messages
        so commit needs to happen frequently (make sure to decrease
        :setting:`broker_commit_every`).
        """

    @sections.Stream.setting(
        params.Seconds,
        version_introduced="1.3",
        version_changed={
            "1.5.3": "Disabled by default.",
        },
        env_name="STREAM_RECOVERY_DELAY",
        default=0.0,
    )
    def stream_recovery_delay(self) -> float:
        """Stream recovery delayl

        Number of seconds to sleep before continuing after rebalance.
        We wait for a bit to allow for more nodes to join/leave before
        starting recovery tables and then processing streams. This to minimize
        the chance of errors rebalancing loops.
        """

    @sections.Stream.setting(
        params.Bool,
        env_name="STREAM_WAIT_EMPTY",
        default=True,
    )
    def stream_wait_empty(self) -> bool:
        """Stream wait empty.

        This setting controls whether the worker should wait for the
        currently processing task in an agent to complete before
        rebalancing or shutting down.

        On rebalance/shut down we clear the stream buffers. Those events
        will be reprocessed after the rebalance anyway, but we may have
        already started processing one event in every agent, and if we
        rebalance we will process that event again.

        By default we will wait for the currently active tasks, but if your
        streams are idempotent you can disable it using this setting.
        """

    @sections.Common.setting(
        params.URL,
        env_name="APP_STORE",
        default="memory://",
    )
    def store(self) -> URL:
        """Table storage backend URL.

        The backend used for table storage.

        Tables are stored in-memory by default, but you should
        not use the ``memory://`` store in production.

        In production, a persistent table store, such as ``rocksdb://`` is
        preferred.
        """

    @sections.Table.setting(
        params.Seconds,
        env_name="TABLE_CLEANUP_INTERVAL",
        default=30.0,
    )
    def table_cleanup_interval(self) -> float:
        """Table cleanup interval.

        How often we cleanup tables to remove expired entries.
        """

    @sections.Table.setting(
        params.UnsignedInt,
        version_introduced="1.7",
        env_name="TABLE_KEY_INDEX_SIZE",
        default=1000,
    )
    def table_key_index_size(self) -> int:
        """Table key index size.

        Tables keep a cache of key to partition number to speed up
        table lookups.

        This setting configures the maximum size of that cache.
        """

    @sections.Table.setting(
        params.UnsignedInt,
        env_name="TABLE_STANDBY_REPLICAS",
        default=1,
    )
    def table_standby_replicas(self) -> int:
        """Table standby replicas.

        The number of standby replicas for each table.
        """

    @sections.Topic.setting(
        params.Bool,
        version_introduced="1.5",
        env_name="TOPIC_ALLOW_DECLARE",
        default=True,
    )
    def topic_allow_declare(self) -> bool:
        """Allow creating new topics.

        This setting disables the creation of internal topics.

        Faust will only create topics that it considers to be fully
        owned and managed, such as intermediate repartition topics,
        table changelog topics etc.

        Some Kafka managers does not allow services to create topics,
        in that case you should set this to :const:`False`.
        """

    @sections.Topic.setting(
        params.Bool,
        version_introduced="1.7",
        env_name="TOPIC_DISABLE_LEADER",
        default=False,
    )
    def topic_disable_leader(self) -> bool:
        """Disable leader election topic.

        This setting disables the creation of the leader election topic.

        If you're not using the ``on_leader=True`` argument to task/timer/etc.,
        decorators then use this setting to disable creation of the topic.
        """

    @sections.Topic.setting(
        params.UnsignedInt,
        env_name="TOPIC_PARTITIONS",
        default=8,
    )
    def topic_partitions(self) -> int:
        """Topic partitions.

        Default number of partitions for new topics.

        .. note::

            This defines the maximum number of workers we could distribute the
            workload of the application (also sometimes referred as the
            sharding factor of the application).
        """

    @sections.Topic.setting(
        params.UnsignedInt,
        env_name="TOPIC_REPLICATION_FACTOR",
        default=1,
    )
    def topic_replication_factor(self) -> int:
        """Topic replication factor.

        The default replication factor for topics created by the application.

        .. note::

            Generally this should be the same as the configured
            replication factor for your Kafka cluster.
        """

    @sections.Common.setting(
        params.URL,
        version_introduced="1.2",
        env_name="CACHE_URL",
        default="memory://",
    )
    def cache(self) -> URL:
        """Cache backend URL.

        Optional backend used for Memcached-style caching.
        URL can be:

        + ``redis://host``
        + ``rediscluster://host``, or
        + ``memory://``.
        """

    @sections.WebServer.setting(
        params.URL,
        version_introduced="1.2",
        default="aiohttp://",
    )
    def web(self) -> URL:
        """Web server driver to use."""

    @sections.WebServer.setting(
        params.Str,
        version_introduced="1.2",
        env_name="WEB_BIND",
        default=WEB_BIND,
        related_cli_options={"faust worker": ["--web-bind"]},
    )
    def web_bind(self) -> str:
        """Web network interface binding mask.

        The IP network address mask that decides what interfaces
        the web server will bind to.

        By default this will bind to all interfaces.

        This option is usually set by :option:`faust worker --web-bind`,
        not by passing it as a keyword argument to :class:`app`.
        """

    @sections.WebServer.setting(
        params.Dict[ResourceOptions],
        version_introduced="1.5",
    )
    def web_cors_options(self) -> Mapping[str, ResourceOptions]:
        """Cross Origin Resource Sharing options.

        Enable `Cross-Origin Resource Sharing`_ options for all web views
        in the internal web server.

        This should be specified as a dictionary of
        URLs to :class:`~faust.web.ResourceOptions`:

        .. sourcecode:: python

            app = App(..., web_cors_options={
                'http://foo.example.com': ResourceOptions(
                    allow_credentials=True,
                    allow_methods='*'k,
                )
            })

        Individual views may override the CORS options used as
        arguments to to ``@app.page`` and ``blueprint.route``.

        .. seealso::

            :pypi:`aiohttp_cors`: https://github.com/aio-libs/aiohttp-cors

        .. _`Cross-Origin Resource Sharing`:
            https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS
        """

    @sections.WebServer.setting(
        params.Bool,
        version_introduced="1.2",
        env_name="APP_WEB_ENABLED",
        default=True,
        related_cli_options={"faust worker": ["--with-web"]},
    )
    def web_enabled(self) -> bool:
        """Enable/disable internal web server.

        Enable web server and other web components.

        This option can also be set using :option:`faust worker --without-web`.
        """

    @sections.WebServer.setting(
        params.Str,
        version_introduced="1.2",
        env_name="WEB_HOST",
        default_template="{conf.NODE_HOSTNAME}",
        related_cli_options={"faust worker": ["--web-host"]},
    )
    def web_host(self) -> str:
        """Web server host name.

        Hostname used to access this web server, used for generating
        the :setting:`canonical_url` setting.

        This option is usually set by :option:`faust worker --web-host`,
        not by passing it as a keyword argument to :class:`app`.
        """

    @sections.WebServer.setting(
        params.Bool,
        version_introduced="1.5",
        default=False,
    )
    def web_in_thread(self) -> bool:
        """Run the web server in a separate thread.

        Use this if you have a large value for
        :setting:`stream_buffer_maxsize` and want the web server
        to be responsive when the worker is otherwise busy processing streams.

        .. note::

            Running the web server in a separate thread means web views
            and agents will not share the same event loop.
        """

    @sections.WebServer.setting(
        params.Port,
        version_introduced="1.2",
        env_name="WEB_PORT",
        default=WEB_PORT,
        related_cli_options={"faust worker": ["--web-port"]},
    )
    def web_port(self) -> int:
        """Web server port.

        A port number between 1024 and 65535 to use for the web server.

        This option is usually set by :option:`faust worker --web-port`,
        not by passing it as a keyword argument to :class:`app`.
        """

    @sections.WebServer.setting(
        params.SSLContext,
        version_introduced="0.5.0",
        default=None,
    )
    def web_ssl_context(self) -> ssl.SSLContext:
        """Web server SSL configuration.

        See :setting:`credentials`.
        """

    @sections.WebServer.setting(
        params.URL,
        version_introduced="1.2",
        default=WEB_TRANSPORT,
        related_cli_options={"faust worker": ["--web-transport"]},
    )
    def web_transport(self) -> URL:
        """Network transport used for the web server.

        Default is to use TCP, but this setting also enables you to use
        Unix domainN sockets.  To use domain sockets specify an URL including
        the path to the file you want to create like this:

        .. sourcecode:: text

            unix:///tmp/server.sock

        This will create a new domain socket available
        in :file:`/tmp/server.sock`.
        """

    @sections.WebServer.setting(
        params.URL,
        default_template="http://{conf.web_host}:{conf.web_port}",
        env_name="NODE_CANONICAL_URL",
        related_cli_options={
            "faust worker": ["--web-host", "--web-port"],
        },
        related_settings=[web_host, web_port],
    )
    def canonical_url(self) -> URL:
        """Node specific canonical URL.

        You shouldn't have to set this manually.

        The canonical URL defines how to reach the web server on a running
        worker node, and is usually set by combining the
        :setting:`web_host` and :setting:`web_port` settings.
        """

    @sections.Worker.setting(
        params.Bool,
        env_name="WORKER_REDIRECT_STDOUTS",
        default=True,
    )
    def worker_redirect_stdouts(self) -> bool:
        """Redirecting standard outputs.

        Enable to have the worker redirect output to :data:`sys.stdout` and
        :data:`sys.stderr` to the Python logging system.

        Enabled by default.
        """

    @sections.Worker.setting(
        params.Severity,
        env_name="WORKER_REDIRECT_STDOUTS_LEVEL",
        default="WARN",
    )
    def worker_redirect_stdouts_level(self) -> Severity:
        """Level used when redirecting standard outputs.

        The logging level to use when redirect STDOUT/STDERR to logging.
        """

    @sections.Extension.setting(
        params.Symbol(Type[AgentT]),
        default="faust:Agent",
    )
    def Agent(self) -> Type[AgentT]:
        """Agent class type.

        The :class:`~faust.Agent` class to use for agents, or the
        fully-qualified path to one (supported by
        :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            class MyAgent(faust.Agent):
                ...

            app = App(..., Agent=MyAgent)

        Example using the string path to a class::

            app = App(..., Agent='myproj.agents.Agent')
        """

    @sections.Consumer.setting(
        params.Symbol(Type[SchedulingStrategyT]),
        version_introduced="1.5",
        default="faust.transport.utils:DefaultSchedulingStrategy",
    )
    def ConsumerScheduler(self) -> Type[SchedulingStrategyT]:
        """Consumer scheduler class.

        A strategy which dictates the priority of topics and
        partitions for incoming records.
        The default strategy does first round-robin over topics and then
        round-robin over partitions.

        Example using a class::

            class MySchedulingStrategy(DefaultSchedulingStrategy):
                ...

            app = App(..., ConsumerScheduler=MySchedulingStrategy)

        Example using the string path to a class::

            app = App(..., ConsumerScheduler='myproj.MySchedulingStrategy')
        """

    @sections.Extension.setting(
        params.Symbol(Type[EventT]),
        default="faust:Event",
    )
    def Event(self) -> Type[EventT]:
        """Event class type.

        The :class:`~faust.Event` class to use for creating new event objects,
        or the fully-qualified path to one (supported by
        :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            class MyBaseEvent(faust.Event):
                ...

            app = App(..., Event=MyBaseEvent)

        Example using the string path to a class::

            app = App(..., Event='myproj.events.Event')
        """

    @sections.Extension.setting(
        params.Symbol(Type[SchemaT]),
        default="faust:Schema",
    )
    def Schema(self) -> Type[SchemaT]:
        """Schema class type.

        The :class:`~faust.Schema` class to use as the default
        schema type when no schema specified. or the fully-qualified
        path to one (supported by :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            class MyBaseSchema(faust.Schema):
                ...

            app = App(..., Schema=MyBaseSchema)

        Example using the string path to a class::

            app = App(..., Schema='myproj.schemas.Schema')
        """

    @sections.Extension.setting(
        params.Symbol(Type[StreamT]),
        default="faust:Stream",
    )
    def Stream(self) -> Type[StreamT]:
        """Stream class type.

        The :class:`~faust.Stream` class to use for streams, or the
        fully-qualified path to one (supported by
        :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            class MyBaseStream(faust.Stream):
                ...

            app = App(..., Stream=MyBaseStream)

        Example using the string path to a class::

            app = App(..., Stream='myproj.streams.Stream')
        """

    @sections.Extension.setting(
        params.Symbol(Type[TableT]),
        default="faust:Table",
    )
    def Table(self) -> Type[TableT]:
        """Table class type.

        The :class:`~faust.Table` class to use for tables, or the
        fully-qualified path to one (supported by
        :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            class MyBaseTable(faust.Table):
                ...

            app = App(..., Table=MyBaseTable)

        Example using the string path to a class::

            app = App(..., Table='myproj.tables.Table')
        """

    @sections.Extension.setting(
        params.Symbol(Type[TableT]),
        default="faust:SetTable",
    )
    def SetTable(self) -> Type[TableT]:
        """SetTable extension table.

        The :class:`~faust.SetTable` class to use for table-of-set tables,
        or the fully-qualified path to one (supported
        by :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            class MySetTable(faust.SetTable):
                ...

            app = App(..., Table=MySetTable)

        Example using the string path to a class::

            app = App(..., Table='myproj.tables.MySetTable')
        """

    @sections.Extension.setting(
        params.Symbol(Type[GlobalTableT]),
        default="faust:GlobalTable",
    )
    def GlobalTable(self) -> Type[GlobalTableT]:
        """GlobalTable class type.

        The :class:`~faust.GlobalTable` class to use for tables,
        or the fully-qualified path to one (supported by
        :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            class MyBaseGlobalTable(faust.GlobalTable):
                ...

            app = App(..., GlobalTable=MyBaseGlobalTable)

        Example using the string path to a class::

            app = App(..., GlobalTable='myproj.tables.GlobalTable')
        """

    @sections.Extension.setting(
        params.Symbol(Type[GlobalTableT]),
        default="faust:SetGlobalTable",
    )
    def SetGlobalTable(self) -> Type[GlobalTableT]:
        """SetGlobalTable class type.

        The :class:`~faust.SetGlobalTable` class to use for tables, or the
        fully-qualified path to one (supported by
        :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            class MyBaseSetGlobalTable(faust.SetGlobalTable):
                ...

            app = App(..., SetGlobalTable=MyBaseGlobalSetTable)

        Example using the string path to a class::

            app = App(..., SetGlobalTable='myproj.tables.SetGlobalTable')
        """

    @sections.Extension.setting(
        params.Symbol(Type[TableManagerT]),
        default="faust.tables:TableManager",
    )
    def TableManager(self) -> Type[TableManagerT]:
        """Table manager class type.

        The :class:`~faust.tables.TableManager` used for managing tables,
        or the fully-qualified path to one (supported by
        :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            from faust.tables import TableManager

            class MyTableManager(TableManager):
                ...

            app = App(..., TableManager=MyTableManager)

        Example using the string path to a class::

            app = App(..., TableManager='myproj.tables.TableManager')
        """

    @sections.Extension.setting(
        params.Symbol(Type[RegistryT]),
        default="faust.serializers:Registry",
    )
    def Serializers(self) -> Type[RegistryT]:
        """Serializer registry class type.

        The :class:`~faust.serializers.Registry` class used for
        serializing/deserializing messages; or the fully-qualified path
        to one (supported by :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            from faust.serialiers import Registry

            class MyRegistry(Registry):
                ...

            app = App(..., Serializers=MyRegistry)

        Example using the string path to a class::

            app = App(..., Serializers='myproj.serializers.Registry')
        """

    @sections.Extension.setting(
        params.Symbol(Type[_WorkerT]),
        default="faust.worker:Worker",
    )
    def Worker(self) -> Type[_WorkerT]:
        """Worker class type.

        The :class:`~faust.Worker` class used for starting a worker
        for this app; or the fully-qualified path
        to one (supported by :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            import faust

            class MyWorker(faust.Worker):
                ...

            app = faust.App(..., Worker=Worker)

        Example using the string path to a class::

            app = faust.App(..., Worker='myproj.workers.Worker')
        """

    @sections.Extension.setting(
        params.Symbol(Type[PartitionAssignorT]),
        default="faust.assignor:PartitionAssignor",
    )
    def PartitionAssignor(self) -> Type[PartitionAssignorT]:
        """Partition assignor class type.

        The :class:`~faust.assignor.PartitionAssignor` class used for assigning
        topic partitions to worker instances; or the fully-qualified path
        to one (supported by :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            from faust.assignor import PartitionAssignor

            class MyPartitionAssignor(PartitionAssignor):
                ...

            app = App(..., PartitionAssignor=PartitionAssignor)

        Example using the string path to a class::

            app = App(..., Worker='myproj.assignor.PartitionAssignor')
        """

    @sections.Extension.setting(
        params.Symbol(Type[LeaderAssignorT]),
        default="faust.assignor:LeaderAssignor",
    )
    def LeaderAssignor(self) -> Type[LeaderAssignorT]:
        """Leader assignor class type.

        The :class:`~faust.assignor.LeaderAssignor` class used for assigning
        a master Faust instance for the app; or the fully-qualified path
        to one (supported by :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            from faust.assignor import LeaderAssignor

            class MyLeaderAssignor(LeaderAssignor):
                ...

            app = App(..., LeaderAssignor=LeaderAssignor)

        Example using the string path to a class::

            app = App(..., Worker='myproj.assignor.LeaderAssignor')
        """

    @sections.Extension.setting(
        params.Symbol(Type[RouterT]),
        default="faust.app.router:Router",
    )
    def Router(self) -> Type[RouterT]:
        """Router class type.

        The :class:`~faust.router.Router` class used for routing requests
        to a worker instance having the partition for a specific
        key (e.g. table key); or the fully-qualified path to one
        (supported by :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            from faust.router import Router

            class MyRouter(Router):
                ...

            app = App(..., Router=Router)

        Example using the string path to a class::

            app = App(..., Router='myproj.routers.Router')
        """

    @sections.Extension.setting(
        params.Symbol(Type[TopicT]),
        default="faust:Topic",
    )
    def Topic(self) -> Type[TopicT]:
        """Topic class type.

        The :class:`~faust.Topic` class used for defining new topics; or the
        fully-qualified path to one (supported by
        :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            import faust

            class MyTopic(faust.Topic):
                ...

            app = faust.App(..., Topic=MyTopic)

        Example using the string path to a class::

            app = faust.App(..., Topic='myproj.topics.Topic')
        """

    @sections.Extension.setting(
        params.Symbol(Type[HttpClientT]),
        default="aiohttp.client:ClientSession",
    )
    def HttpClient(self) -> Type[HttpClientT]:
        """Http client class type

        The :class:`aiohttp.client.ClientSession` class used as
        a HTTP client; or the fully-qualified path to one (supported by
        :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            import faust
            from aiohttp.client import ClientSession

            class HttpClient(ClientSession):
                ...

            app = faust.App(..., HttpClient=HttpClient)

        Example using the string path to a class::

            app = faust.App(..., HttpClient='myproj.http.HttpClient')
        """

    @sections.Extension.setting(
        params.Symbol(Type[SensorT]),
        default="faust.sensors:Monitor",
    )
    def Monitor(self) -> Type[SensorT]:
        """Monitor sensor class type.

        The :class:`~faust.sensors.Monitor` class as the main sensor
        gathering statistics for the application; or the
        fully-qualified path to one (supported by
        :func:`~mode.utils.imports.symbol_by_name`).

        Example using a class::

            import faust
            from faust.sensors import Monitor

            class MyMonitor(Monitor):
                ...

            app = faust.App(..., Monitor=MyMonitor)

        Example using the string path to a class::

            app = faust.App(..., Monitor='myproj.monitors.Monitor')
        """

    @sections.Stream.setting(
        params.Bool,
        default=True,
        version_deprecated="1.0",
        deprecation_reason="no longer has any effect",
    )
    def stream_ack_cancelled_tasks(self) -> bool:
        """Deprecated setting has no effect."""

    @sections.Stream.setting(
        params.Bool,
        default=True,
        version_deprecated="1.0",
        deprecation_reason="no longer has any effect",
    )
    def stream_ack_exceptions(self) -> bool:
        """Deprecated setting has no effect."""

    @sections.Producer.setting(
        params.UnsignedInt,
        env_name="PRODUCER_LINGER_MS",
        version_deprecated="1.11",
        deprecation_reason="use producer_linger in seconds instead.",
        default=0,
    )
    def producer_linger_ms(self) -> int:
        """Deprecated setting, please use :setting:`producer_linger` instead.

        This used to be provided as milliseconds, the new setting
        uses seconds.
        """

    @sections.Common.setting(
        params.URL,
        default=None,
        version_deprecated=1.0,
        deprecation_reason='Please use "broker" setting instead',
    )
    def url(self) -> URL:
        """Backward compatibility alias to :setting:`broker`."""



"""Agents."""
from .agent import Agent, AgentFun, AgentT, SinkT, current_agent
from .manager import AgentManager, AgentManagerT
from .replies import ReplyConsumer

__all__ = [
    "Agent",
    "AgentFun",
    "AgentT",
    "SinkT",
    "AgentManager",
    "AgentManagerT",
    "ReplyConsumer",
    "current_agent",
]
